import pandas as pd

class denseDF2DBPlus:
    """
        This class create Data Base from DataFrame.

        Attribute:
        ----------
        inputDF : pandas.DataFrame
            It is dense DataFrame which was transposed.
        thresholdConditionDF : pandas.DataFrame
            It has threshold value and condition each item.
        tids : list
            It is tids list.
        items : list
            Store the items list
        df : pandas.DataFrame
            It is data frame which is merged inputDF and thresholdConditionDF.
        outputFile : str
            Creation data base output to this outputFile.

        Methods:
        --------
        createDB(outputFile)
            Create transactional data base from dataFrame
        createTDB(outputFile)
            Create temporal dataBase from dataFrame
        createUDB(outputFile)
            Create utility database from dataFrame
        getFileName()
            Return outputFileName.
        """

    def __init__(self, inputDF, thresholdConditionDF):
        self.inputDF = inputDF.T
        self.thresholdConditionDF = thresholdConditionDF
        self.tids = []
        self.items = []
        self.outputFile = ' '
        self.items = list(self.inputDF.index)
        self.tids = list(self.inputDF.columns)
        self.df = pd.merge(self.inputDF, self.thresholdConditionDF, left_index=True, right_index=True)


    def createTransactional(self, outputFile):
        """
        Create transactional data base

        :param outputFile: Write transactional data base into outputFile
        :type outputFile: str
        """

        self.outputFile = outputFile
        with open(outputFile, 'w') as f:
            for tid in self.tids:
                transaction = [item for item in self.items if
                               (self.df.at[item, 'condition'] == '>' and self.df.at[item, tid] > self.df.at[item, 'threshold']) or
                               (self.df.at[item, 'condition'] == '>=' and self.df.at[item, tid] >= self.df.at[item, 'threshold']) or
                               (self.df.at[item, 'condition'] == '<=' and self.df.at[item, tid] <= self.df.at[item, 'threshold']) or
                               (self.df.at[item, 'condition'] == '<' and self.df.at[item, tid] < self.df.at[item, 'threshold']) or
                               (self.df.at[item, 'condition'] == '==' and self.df.at[item, tid] == self.df.at[item, 'threshold']) or
                               (self.df.at[item, 'condition'] == '!=' and self.df.at[item, tid] != self.df.at[item, 'threshold'])]
                if len(transaction) > 1:
                    f.write(f'{transaction[0]}')
                    for item in transaction[1:]:
                        f.write(f',{item}')
                elif len(transaction) == 1:
                    f.write(f'{transaction[0]}')
                else:
                    continue
                f.write('\n')



    def createTemporal(self, outputFile):
        """
        Create temporal data base

        :param outputFile: Write temporal data base into outputFile
        :type outputFile: str
        """

        self.outputFile = outputFile
        with open(outputFile, 'w') as f:
            for tid in self.tids:
                transaction = [item for item in self.items if
                               (self.df.at[item, 'condition'] == '>' and self.df.at[item, tid] > self.df.at[item, 'threshold']) or
                               (self.df.at[item, 'condition'] == '>=' and self.df.at[item, tid] >= self.df.at[item, 'threshold']) or
                               (self.df.at[item, 'condition'] == '<=' and self.df.at[item, tid] <= self.df.at[item, 'threshold']) or
                               (self.df.at[item, 'condition'] == '<' and self.df.at[item, tid] < self.df.at[item, 'threshold']) or
                               (self.df.at[item, 'condition'] == '==' and self.df.at[item, tid] == self.df.at[item, 'threshold']) or
                               (self.df.at[item, 'condition'] == '!=' and self.df.at[item, tid] != self.df.at[item, 'threshold'])]
                if len(transaction) > 1:
                    f.write(f'{tid}')
                    for item in transaction:
                        f.write(f',{item}')
                elif len(transaction) == 1:
                    f.write(f'{tid}')
                    f.write(f',{transaction[0]}')
                else:
                    continue
                f.write('\n')

    def createUtility(self, outputFile):
        """
        Create the utility data base.

        :param outputFile: Write utility data base into outputFile
        :type outputFile: str
        """

        self.outputFile = outputFile
        with open(self.outputFile, 'w') as f:
            for tid in self.tids:
                df = self.inputDF.loc[tid].dropna()
                f.write(f'{df.index[0]}')
                for item in df.index[1:]:
                    f.write(f'\t{item}')
                f.write(f':{df.sum()}:')
                f.write(f'{df.at[df.index[0]]}')
                for item in df.index[1:]:
                    f.write(f'\t{df.at[item]}')
                f.write('\n')

    def getFileName(self):
        """
        return outputFile name

        :return: outputFile name
        """

        return self.outputFile


import sys
import time
import resource
import math
import numpy as np
path=sys.argv[1]
output=sys.argv[2]
itemSup=int(sys.argv[3])
minSup=int(sys.argv[4])
faultTolerance=int(sys.argv[5])
maxPer=int(sys.argv[6])
Vector={}
lno=0
plist=[]
transactions=[]
items=[]
def Per_Sup(tids):
    per=0
    cur=0
    sup=0
    for i in range(len(tids)):
        if tids[i]==1:
            per=max(per,cur-i)
            if per>maxPer:
                return [0,0]
            cur=i
            sup+=1
    per=max(per,lno-cur)
    return [sup,per]
def Count(tids):
    count=0
    for i in tids:
        if i==1:
            count+=1
    return count
         
with open(path,'r') as f:
    for line in f:
        lno+=1
        l=line.split()
        transactions.append(l)
        for i in l[1:]:
            if i not in plist:
                plist.append(i)
for i in transactions:
    for j in plist:
        count=0
        if j in i:
            count=1
        if j in Vector:
            Vector[j].append(count)
        else:
            Vector[j]=[count]
for x,y in Vector.items():
    v=Count(y)
    if v>=itemSup:
        items.append(x)
        
def save(prefix,suffix,tidsetx):
        if(prefix==None):
            prefix=suffix
        else:
            prefix=prefix+suffix
        prefix=list(set(prefix))
        prefix.sort()
        val=Count(tidsetx)
        print(prefix,val)
        
def processEquivalenceClass(prefix,itemsets,tidsets):
        if(len(itemsets)==1):
            i=itemsets[0]
            tidi=tidsets[0]
            save(prefix,[i],tidi)
            return
        for i in range(len(itemsets)):
            itemx=itemsets[i]
            if(itemx==None):
                continue
            tidsetx=tidsets[i]
            classItemsets=[]
            classtidsets=[]
            itemsetx=[itemx]
            for j in range(i+1,len(itemsets)):
                itemj=itemsets[j]
                tidsetj=tidsets[j]
                y=list(np.array(tidsetx) & np.array(tidsetj))
                total=Count(y)
                if total>=itemSup:
                    classItemsets.append(itemj)
                    classtidsets.append(y)
            if(len(classItemsets)>0):
                newprefix=list(set(itemsetx))+prefix
                processEquivalenceClass(newprefix, classItemsets,classtidsets,classItemSets)
            save(prefix,list(set(itemsetx)),itemTidsi)

def Mine(plist):
    for i in range(len(plist)):
            itemx=plist[i]
            tidsetx=Vector[itemx]
            itemsetx=[itemx]
            itemsets=[]
            itemSets=[]
            tidsets=[]
            for j in range(i+1,len(plist)):
                itemj=plist[j]
                tidsetj=Vector[itemj]
                y1=list(np.array(tidsetx) & np.array(tidsetj))
                total=Count(y1)
                if total>=itemSup:
                    itemsets.append(itemj)
                    tidsets.append(y1)
            if(len(itemsets)>0):
                processEquivalenceClass(itemsetx,itemsets,tidsets)
            save(None,itemsetx,tidsetx)
    
               
Mine(items)



import pandas as pd

class sparseDF2DB:
    """
    This class create Data Base from DataFrame.

    Attribute:
    ----------
    inputDF : pandas.DataFrame
        It is sparse DataFrame
    condition : str
        It is condition to judge the value in dataframe
    thresholdValue : int or float:
        User defined value.
    df : pandas.DataFrame
        It is data frame to create data base.
    outputFile : str
        Creation data base output to this outputFile.

    Methods:
    --------
    createDB(outputFile)
        Create transactional data base from dataFrame
    createTDB(outputFile)
        Create temporal dataBase from dataFrame
    createUDB(outputFile)
        Create utility data base from dataFrame
    getFileName()
        Return outputFileName.
    """

    def __init__(self, inputDF, condition, thresholdValue):
        self.inputDF = inputDF
        self.condition = condition
        self.thresholdValue = thresholdValue
        self.outputFile = ''
        if self.condition == '>':
            self.df = self.inputDF.query(f'value > {self.thresholdValue}')
        elif self.condition == '>=':
            self.df = self.inputDF.query(f'value >= {self.thresholdValue}')
        elif self.condition == '<=':
            self.df = self.inputDF.query(f'value <= {self.thresholdValue}')
        elif self.condition == '<':
            self.df = self.inputDF.query(f'value < {self.thresholdValue}')
        else:
            print('Condition error')
        self.df = self.df.drop(columns='value')
        self.df = self.df.groupby('tid')['item'].apply(list)

    def createTransactional(self, outputFile):
        """
        Create transactional data base
        :param outputFile: Write transactional data base into outputFile
        :type outputFile: str
        """
        self.outputFile = outputFile
        with open(self.outputFile, 'w') as f:
            for line in self.df:
                f.write(f'{line[0]}')
                for item in line[1:]:
                    f.write(f',{item}')
                f.write('\n')

    def createTemporal(self, outputFile):
        """
        Create temporal data base

        :param outputFile: Write temporal data base into outputFile
        :type outputFile: str
        """

        self.outputFile = outputFile
        with open(self.outputFile, 'w') as f:
            for tid in self.df.index:
                f.write(f'{tid}')
                for item in self.df[tid]:
                    f.write(f',{item}')
                f.write('\n')

    def createUtility(self, outputFile):
        """
        Create the utility data base.

        :param outputFile: Write utility data base into outputFile
        :type outputFile: str
        """

        self.outputFile = outputFile
        items = self.inputDF.groupby(level=0)['item'].apply(list)
        values = self.inputDF.groupby(level=0)['value'].apply(list)
        sums = self.inputDF.groupby(level=0)['value'].sum()
        index = list(items.index)
        with open(self.outputFile, 'w') as f:
            for tid in index:
                f.write(f'{items[tid][0]}')
                for item in items[tid][1:]:
                    f.write(f'\t{item}')
                f.write(f':{sums[tid]}:')
                f.write(f'{values[tid][0]}')
                for value in values[tid][1:]:
                    f.write(f'\t{value}')
                f.write('\n')

    def getFileName(self):
        """
        return outputFile name
        :return: outputFile name
        """
        return self.outputFile

if __name__ == '__main__':
    DF = createSparseDF('sparseDF.csv')
    obj = sparseDF2DB(DF.getDF(), '>=', 2)
    obj.createDB('testTransactional.csv')
    transactionalDB = obj.getFileName()
    print(transactionalDB)




from PAMI.extras.DF2DB.denseDF2DB import *
from PAMI.extras.DF2DB.sparseDF2DB import *

class DF2DB:
    """
        This class create database from DataFrame. Threshold values and conditions are defined to all item.

        Attribute:
        ----------
        inputDF : pandas.DataFrame
            It is sparse or dense DataFrame
        thresholdValue : int or float
            It is threshold value of all item
        condition : str
            It is condition of all item
        DFtype : str
            It is DataFrame type. It should be sparse or dense. Default DF is sparse.

        Nethods:
        --------
        getDB(outputFile)
            Create transactional database from DataFrame and store into outputFile
        getTDB(outputFile)
            Create temporal database from DataFrame and store into outputFile
        getUDB(outputFile)
            Create utility database from DataFrame and store into outputFile
        """
    def __init__(self, inputDF, thresholdValue, condition, DFtype='sparse'):
        self.inputDF = inputDF
        self.thresholdValue = thresholdValue
        self.condition = condition
        self.DFtype = DFtype.lower()
        if DFtype == 'sparse':
            self.DF2DB = sparseDF2DB(self.inputDF, self.condition, self.thresholdValue)
        elif DFtype == 'dense':
            self.DF2DB = denseDF2DB(self.inputDF, self.condition, self.thresholdValue)
        else:
            raise Exception('DF type should be sparse or dense')

    def getTransactional(self, outputFile):
        """
        create transactional database and return outputFileName
        :param outputFile: file name or path to store database
        :type outputFile: str
        :return: outputFile name
        """
        self.DF2DB.createTransactional(outputFile)
        return self.DF2DB.getFileName()

    def getTemporal(self, outputFile):
        """
        create temporal database and return outputFile name
        :param outputFile: file name or path to store database
        :type outputFile: str
        :return: outputFile name
        """
        self.DF2DB.createTemporal(outputFile)
        return self.DF2DB.getFileName()

    def getUtility(self, outputFile):
        """
        create utility database and return outputFile name
        :param outputFile:  file name or path to store database
        :type outputFile: str
        :return: outputFile name
        """
        self.DF2DB.createUtility(outputFile)
        return self.DF2DB.getFileName()



#  Copyright (C)  2021 Rage Uday Kiran
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.


from PAMI.partialPeriodicPattern.basic import abstract as _ab


class PPP_ECLAT(_ab._partialPeriodicPatterns):
    """
    3pEclat is the fundamental approach to mine the partial periodic frequent patterns.

    Reference:

    Parameters:
    ----------
        self.iFile : file
            Name of the Input file or path of the input file
        self. oFile : file
            Name of the output file or path of the output file
        periodicSupport: float or int or str
            The user can specify periodicSupport either in count or proportion of database size.
            If the program detects the data type of periodicSupport is integer, then it treats periodicSupport is expressed in count.
            Otherwise, it will be treated as float.
            Example: periodicSupport=10 will be treated as integer, while periodicSupport=10.0 will be treated as float
        period: float or int or str
            The user can specify period either in count or proportion of database size.
            If the program detects the data type of period is integer, then it treats period is expressed in count.
            Otherwise, it will be treated as float.
            Example: period=10 will be treated as integer, while period=10.0 will be treated as float
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
            However, the users can override their default separator.
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        Database : list
            To store the transactions of a database in list
        mapSupport : Dictionary
            To maintain the information of item and their frequency
        lno : int
            it represents the total no of transactions
        tree : class
            it represents the Tree class
        finalPatterns : dict
            it represents to store the patterns
        tidList : dict
            stores the timestamps of an item

    Methods:
    -------
        startMine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        savePatterns(oFile)
            Complete set of frequent patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        creatingOneitemSets()
            Scan the database and store the items with their timestamps which are periodic frequent
        getPeriodAndSupport()
            Calculates the support and period for a list of timestamps.
        Generation()
            Used to implement prefix class equivalence method to generate the periodic patterns recursively

    Executing the code on terminal:
    -------

        Format: python3 PPP_ECLAT.py <inputFile> <outputFile> <periodicSupport> <period>

        Examples: python3 PPP_ECLAT.py sampleDB.txt patterns.txt 0.3 0.4   (periodicSupport and period will be considered in percentage of database transactions)

                  python3 threePEeclat.py sampleDB.txt patterns.txt 3 4     (periodicSupport and period will be considered in support count or frequency)


    Sample run of importing the code:
    -------------------

        from PAMI.periodicFrequentPattern.basic import PPP_ECLAT as alg

        obj = alg.PPP_ECLAT(iFile, periodicSupport,period)

        obj.startMine()

        Patterns = obj.getPatterns()

        print("Total number of partial periodic patterns:", len(Patterns))

        obj.savePatterns(oFile)

        Df = obj.getPatternsAsDataFrame()

        memUSS = obj.getMemoryUSS()

        print("Total Memory in USS:", memUSS)

        memRSS = obj.getMemoryRSS()

        print("Total Memory in RSS", memRSS)

        run = obj.getRuntime()

        print("Total ExecutionTime in seconds:", run)

        Credits:
        -------

        The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.\n



        """
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _mapSupport = {}
    _itemsetCount = 0
    _writer = None
    _periodicSupport = str()
    _period = str()
    _tidList = {}
    _lno = 0
    _Database = []

    def _convert(self, value):
        """
        To convert the given user specified value

        :param value: user specified value

        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value

    def _getPeriodicSupport(self, timeStamps):
        """
            calculates the support and periodicity with list of timestamps

            :param timeStamps : timestamps of a pattern

            :type timeStamps : list
        """
        timeStamps.sort()
        per = 0
        for i in range(len(timeStamps) - 1):
            j = i + 1
            if abs(timeStamps[j] - timeStamps[i]) <= self._period:
                per += 1
        return per

    def _creatingItemSets(self):
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            data, tids = [], []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'TS' in i:
                tids = self._iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self._iFile['Transactions'].tolist()
            for i in range(len(data)):
                tr = [tids[i][0]]
                tr = tr + data[i]
                self._Database.append(tr)
            self._lno = len(self._Database)

        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    self._lno += 1
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            self._lno += 1
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def _creatingOneitemSets(self):
        """
           Scans the Temporal database / Input file and stores the 1-length partial-periodic patterns.
        """
        plist = []
        self._tidList = {}
        self._mapSupport = {}
        self._period = self._convert(self._period)
        for line in self._Database:
            s = line
            n = int(s[0])
            for i in range(1, len(s)):
                si = s[i]
                if self._mapSupport.get(si) is None:
                    self._mapSupport[si] = [0, n]
                    self._tidList[si] = [n]
                else:
                    lp = n - self._mapSupport[si][1]
                    if lp <= self._period:
                        self._mapSupport[si][0] += 1
                    self._mapSupport[si][1] = n
                    self._tidList[si].append(n)
        self._periodicSupport = self._convert(self._periodicSupport)
        self._mapSupport = {k: v[0] for k, v in self._mapSupport.items() if v[0] >= self._periodicSupport}
        plist = [key for key, value in sorted(self._mapSupport.items(), key=lambda x: x[1], reverse=True)]
        return plist

    def _save(self, prefix, suffix, tidSetX):
        """
            saves the patterns that satisfy the partial periodic property.

            :param prefix: the prefix of a pattern

            :type prefix: list

            :param suffix : the suffix of a patterns

            :type suffix : list

            :param tidSetX : the timestamp of a patterns

            :type tidSetX : list


        """
        if prefix is None:
            prefix = suffix
        else:
            prefix = prefix + suffix
        val = self._getPeriodicSupport(tidSetX)
        if val >= self._periodicSupport:
            sample = str()
            for i in prefix:
                sample = sample + i + " "
            self._finalPatterns[sample] = val

    def _Generation(self, prefix, itemSets, tidSets):
        """
            Generates the patterns following Equivalence-class methods

            :param prefix :  main equivalence prefix

            :type prefix : partial-periodic item or pattern

            :param itemSets : patterns which are items combined with prefix and satisfying the periodicity
                            and partial property with their timestamps

            :type itemSets : list

            :param tidSets : timestamps of the items in the argument itemSets

            :type tidSets : list


                    """
        if len(itemSets) == 1:
            i = itemSets[0]
            tidi = tidSets[0]
            self._save(prefix, [i], tidi)
            return
        for i in range(len(itemSets)):
            itemI = itemSets[i]
            if itemI is None:
                continue
            tidSetX = tidSets[i]
            classItemSets = []
            classTidSets = []
            itemSetX = [itemI]
            for j in range(i + 1, len(itemSets)):
                itemJ = itemSets[j]
                tidSetJ = tidSets[j]
                y = list(set(tidSetX).intersection(tidSetJ))
                val = self._getPeriodicSupport(y)
                if val >= self._periodicSupport:
                    classItemSets.append(itemJ)
                    classTidSets.append(y)
            newprefix = list(set(itemSetX)) + prefix
            self._Generation(newprefix, classItemSets, classTidSets)
            self._save(prefix, list(set(itemSetX)), tidSetX)

    def startMine(self):
        """
            Main program start with extracting the periodic frequent items from the database and
            performs prefix equivalence to form the combinations and generates partial-periodic patterns.

        """
        self._startTime = _ab._time.time()
        self._creatingItemSets()
        plist = self._creatingOneitemSets()
        self._finalPatterns = {}
        for i in range(len(plist)):
            itemI = plist[i]
            tidSetX = self._tidList[itemI]
            itemSetX = [itemI]
            itemSets = []
            tidSets = []
            for j in range(i + 1, len(plist)):
                itemJ = plist[j]
                tidSetJ = self._tidList[itemJ]
                y1 = list(set(tidSetX).intersection(tidSetJ))
                val = self._getPeriodicSupport(y1)
                if val >= self._periodicSupport:
                    itemSets.append(itemJ)
                    tidSets.append(y1)
            self._Generation(itemSetX, itemSets, tidSets)
            self._save(None, itemSetX, tidSetX)
        print("Partial Periodic Frequent patterns were generated successfully using 3PEclat algorithm")
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryRSS = float()
        self._memoryUSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process

        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process

        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process

        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe

        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b])
            dataframe = _ab._pd.DataFrame(data, columns=['Patterns', 'periodicSupport'])
        return dataframe

    def savePatterns(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file

        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns

        :rtype: dict
        """
        return self._finalPatterns


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:
            _ap = PPP_ECLAT(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:
            _ap = PPP_ECLAT(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        _ap.startMine()
        _Patterns = _ap.getPatterns()
        print("Total number of Partial Periodic Patterns:", len(_Patterns))
        _ap.savePatterns(_ab._sys.argv[2])
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
    else:
        l = [0.001, 0.002, 0.003, 0.004, 0.005]
        for i in l:
            ap = PPP_ECLAT('/Users/Likhitha/Downloads/Datasets/BMS1_itemset_mining.txt', i, 100, ' ')
            ap.startMine()
            Patterns = ap.getPatterns()
            print("Total number of  Patterns:", len(Patterns))
            ap.savePatterns('/Users/Likhitha/Downloads/output')
            memUSS = ap.getMemoryUSS()
            print("Total Memory in USS:", memUSS)
            memRSS = ap.getMemoryRSS()
            print("Total Memory in RSS", memRSS)
            run = ap.getRuntime()
            print("Total ExecutionTime in ms:", run)
        print("Error! The number of input parameters do not match the total number of parameters provided")



#  Copyright (C)  2021 Rage Uday Kiran
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from PAMI.weightedFrequentRegularPattern import abstract as _fp

_WS = str()
_regularity = str()
_lno = int()
_weights = {}
_wf = {}
_fp._sys.setrecursionlimit(20000)


class _Node:
    """
        A class used to represent the node of frequentPatternTree

    Attributes:
    ----------
        itemId: int
            storing item of a node
        counter: int
            To maintain the support of node
        parent: node
            To maintain the parent of node
        children: list
            To maintain the children of node

    Methods:
    -------

        addChild(node)
            Updates the nodes children list and parent for the given node

    """

    def __init__(self, item, children):
        """ Initializing the Node class

        :param item: Storing the item of a node
        :type item: int or None
        :param children: To maintain the children of a node
        :type children: dict
        """

        self.item = item
        self.children = children
        self.parent = None
        self.timeStamps = []

    def addChild(self, node):
        """ To add the children to a node

            :param node: parent node in the tree
        """

        self.children[node.item] = node
        node.parent = self


class _Tree:
    """
    A class used to represent the frequentPatternGrowth tree structure

    Attributes:
    ----------
        root : Node
            The first node of the tree set to Null.
        summaries : dictionary
            Stores the nodes itemId which shares same itemId
        info : dictionary
            frequency of items in the transactions

    Methods:
    -------
        addTransaction(transaction, freq)
            adding items of  transactions into the tree as nodes and freq is the count of nodes
        getFinalConditionalPatterns(node)
            getting the conditional patterns from fp-tree for a node
        getConditionalPatterns(patterns, frequencies)
            sort the patterns by removing the items with lower minSup
        generatePatterns(prefix)
            generating the patterns from fp-tree
    """

    def __init__(self):
        self.root = _Node(None, {})
        self.summaries = {}
        self.info = {}

    def addTransaction(self, transaction, tid):
        """     Adding a transaction into tree

                :param transaction: To represent the complete database
                :type transaction: list
                :param tid: To represent the timestamp of a database
                :type tid: list
                :return: pfp-growth tree
        """

        currentNode = self.root
        for i in range(len(transaction)):
            if transaction[i] not in currentNode.children:
                newNode = _Node(transaction[i], {})
                currentNode.addChild(newNode)
                if transaction[i] in self.summaries:
                    self.summaries[transaction[i]].append(newNode)
                else:
                    self.summaries[transaction[i]] = [newNode]
                currentNode = newNode
            else:
                currentNode = currentNode.children[transaction[i]]
        currentNode.timeStamps = currentNode.timeStamps + tid

    def getConditionalPatterns(self, alpha, pattern):
        """Generates all the conditional patterns of a respective node

            :param alpha: To represent a Node in the tree
            :type alpha: Node
            :param pattern: prefix of the pattern
            :type alpha: list
            :return: A tuple consisting of finalPatterns, conditional pattern base and information
        """
        finalPatterns = []
        finalSets = []
        for i in self.summaries[alpha]:
            set1 = i.timeStamps
            set2 = []
            while i.parent.item is not None:
                set2.append(i.parent.item)
                i = i.parent
            if len(set2) > 0:
                set2.reverse()
                finalPatterns.append(set2)
                finalSets.append(set1)
        finalPatterns, finalSets, info = self.conditionalDatabases(finalPatterns, finalSets, pattern)
        return finalPatterns, finalSets, info

    @staticmethod
    def generateTimeStamps(node):
        """To get the timestamps of a node

        :param node: A node in the tree
        :return: Timestamps of a node
        """

        finalTimeStamps = node.timeStamps
        return finalTimeStamps

    def removeNode(self, nodeValue):
        """ Removing the node from tree

            :param nodeValue: To represent a node in the tree
            :type nodeValue: node
            :return: Tree with their nodes updated with timestamps
        """

        for i in self.summaries[nodeValue]:
            i.parent.timeStamps = i.parent.timeStamps + i.timeStamps
            del i.parent.children[nodeValue]

    def getTimeStamps(self, alpha):
        """ To get all the timestamps of the nodes which share same item name

            :param alpha: Node in a tree
            :return: Timestamps of a  node
        """
        temporary = []
        for i in self.summaries[alpha]:
            temporary += i.timeStamps
        return temporary

    @staticmethod
    def getSupportAndPeriod(timeStamps, pattern):
        """To calculate the periodicity and support

        :param timeStamps: Timestamps of an item set
        :type timeStamps: list
        :param pattern: pattern to evaluate the weighted frequent regular or not
        :type pattern: list
        :return: support, periodicity
        """

        global _WS, _regularity, _lno, _weights
        timeStamps.sort()
        cur = 0
        per = list()
        sup = 0
        for j in range(len(timeStamps)):
            per.append(timeStamps[j] - cur)
            cur = timeStamps[j]
            sup += 1
        per.append(_lno - cur)
        l = int()
        for i in pattern:
            l = l + _weights[i]
        wf = (l / (len(pattern))) * sup
        if len(per) == 0:
            return [0, 0]
        return [sup, max(per), wf]

    def conditionalDatabases(self, conditionalPatterns, conditionalTimeStamps, pattern):
        """ It generates the conditional patterns with periodic-frequent items

            :param conditionalPatterns: conditionalPatterns generated from conditionPattern method of a respective node
            :type conditionalPatterns: list
            :param conditionalTimeStamps: Represents the timestamps of a conditional patterns of a node
            :type conditionalTimeStamps: list
            :param pattern: prefix of the pattern
            :type pattern: list
            :returns: Returns conditional transactions by removing non-periodic and non-frequent items
        """

        global _WS, _regularity
        pat = []
        timeStamps = []
        data1 = {}
        for i in range(len(conditionalPatterns)):
            for j in conditionalPatterns[i]:
                if j in data1:
                    data1[j] = data1[j] + conditionalTimeStamps[i]
                else:
                    data1[j] = conditionalTimeStamps[i]
        updatedDictionary = {}
        for m in data1:
            updatedDictionary[m] = self.getSupportAndPeriod(data1[m], pattern + [m])
        updatedDictionary = {k: v for k, v in updatedDictionary.items() if v[0] >= _WS and v[1] <= _regularity}
        count = 0
        for p in conditionalPatterns:
            p1 = [v for v in p if v in updatedDictionary]
            trans = sorted(p1, key=lambda x: (updatedDictionary.get(x)[0], -x), reverse=True)
            if len(trans) > 0:
                pat.append(trans)
                timeStamps.append(conditionalTimeStamps[count])
            count += 1
        return pat, timeStamps, updatedDictionary

    def generatePatterns(self, prefix):
        """ Generates the patterns

            :param prefix: Forms the combination of items
            :type prefix: list
            :returns: yields patterns with their support and periodicity
        """
        global _WS
        for i in sorted(self.summaries, key=lambda x: (self.info.get(x)[0], -x)):
            pattern = prefix[:]
            pattern.append(i)
            if self.info[i][2] >= _WS:
                yield pattern, self.info[i]
                patterns, timeStamps, info = self.getConditionalPatterns(i, pattern)
                conditionalTree = _Tree()
                conditionalTree.info = info.copy()
                for pat in range(len(patterns)):
                    conditionalTree.addTransaction(patterns[pat], timeStamps[pat])
                if len(patterns) > 0:
                    for q in conditionalTree.generatePatterns(pattern):
                        yield q
            self.removeNode(i)


class WFRIMiner(_fp._weightedFrequentRegularPatterns):
    """
       WFRIMiner is one of the fundamental algorithm to discover weighted frequent regular patterns in a transactional database.
       It stores the database in compressed WFRI-tree decreasing the memory usage and extracts the
       patterns from tree.It employs employs downward closure property to  reduce the search space effectively.

    Reference :
    ---------
           K. Klangwisan and K. Amphawan, "Mining weighted-frequent-regular itemsets from transactional database,"
           2017 9th International Conference on Knowledge and Smart Technology (KST), 2017, pp. 66-71,
           doi: 10.1109/KST.2017.7886090.

    Attributes :
    ----------
        iFile : file
            Input file name or path of the input file
        WS: float or int or str
            The user can specify WS either in count or proportion of database size.
            If the program detects the data type of WS is integer, then it treats WS is expressed in count.
            Otherwise, it will be treated as float.
            Example: WS=10 will be treated as integer, while WS=10.0 will be treated as float
        regularity: float or int or str
            The user can specify regularity either in count or proportion of database size.
            If the program detects the data type of regularity is integer, then it treats regularity is expressed in count.
            Otherwise, it will be treated as float.
            Example: regularity=10 will be treated as integer, while regularity=10.0 will be treated as float
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default separator is tab space or \t.
            However, the users can override their default separator.
        oFile : file
            Name of the output file or the path of the output file
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        Database : list
            To store the transactions of a database in list
        mapSupport : Dictionary
            To maintain the information of item and their frequency
        lno : int
            it represents the total no of transactions
        tree : class
            it represents the Tree class
        finalPatterns : dict
            it represents to store the patterns

    Methods :
    -------
        startMine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        savePatterns(oFile)
            Complete set of frequent patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        creatingItemSets()
            Scans the dataset or dataframes and stores in list format
        frequentOneItem()
            Extracts the one-frequent patterns from transactions

    Executing the code on terminal:
    -------
        Format:
        -------
            python3 WFRIMiner.py <inputFile> <outputFile> <minSup> <regularity>

        Examples:
        ---------
            python3 WFRIMiner.py sampleDB.txt patterns.txt 10.0 5.0  (WS & regularity will be considered in times of WS or regularity and count of database transactions)

            python3 WFRIMiner.py sampleDB.txt patterns.txt 10 5     (WS & regularity will be considered in support count or frequency) (it will consider "\t" as a separator)

            python3 WFRIMiner.py sampleTDB.txt output.txt sampleN.txt '10', '5', ',' (it will consider "," as a separator)


    Sample run of the importing code:
    -----------


        from PAMI.weightedFrequentRegularpattern.basic import WFRIMiner as alg

        obj = alg.WFRIMiner(iFile, WS, regularity)

        obj.startMine()

        weightedFrequentRegularPatterns = obj.getPatterns()

        print("Total number of Frequent Patterns:", len(weightedFrequentRegularPatterns))

        obj.savePatterns(oFile)

        Df = obj.getPatternInDataFrame()

        memUSS = obj.getMemoryUSS()

        print("Total Memory in USS:", memUSS)

        memRSS = obj.getMemoryRSS()

        print("Total Memory in RSS", memRSS)

        run = obj.getRuntime()

        print("Total ExecutionTime in seconds:", run)

    Credits:
    -------
        The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.\n

        """

    _startTime = float()
    _endTime = float()
    _WS = str()
    _regularity = str()
    _weight = {}
    _finalPatterns = {}
    _wFile = " "
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _mapSupport = {}
    _lno = 0
    _tree = _Tree()
    _rank = {}
    _rankDup = {}

    def __init__(self, iFile, _wFile, WS, regularity, sep='\t'):
        super().__init__(iFile, _wFile, WS, regularity, sep)

    def _creatingItemSets(self):
        """
            Storing the complete transactions of the database/input file in a database variable


        """
        self._Database = []
        self._weight = {}
        if isinstance(self._iFile, _fp._pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self._Database = self._iFile['Transactions'].tolist()

        if isinstance(self._wFile, _fp._pd.DataFrame):
            _items, _weights = [], []
            if self._wFile.empty:
                print("its empty..")
            i = self._wFile.columns.values.tolist()
            if 'items' in i:
                _items = self._wFile['items'].tolist()
            if 'weight' in i:
                _weights = self._wFile['weight'].tolist()
            for i in range(len(_items)):
                self._weight[_items[i]] = _weights[i]

            # print(self.Database)
        if isinstance(self._iFile, str):
            if _fp._validators.url(self._iFile):
                data = _fp._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

        if isinstance(self._wFile, str):
            if _fp._validators.url(self._wFile):
                data = _fp._urlopen(self._wFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._weight[temp[0]] = float(temp[1])
            else:
                try:
                    with open(self._wFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._weight[temp[0]] = float(temp[1])
                except IOError:
                    print("File Not Found")
                    quit()

    def _convert(self, value):
        """
        to convert the type of user specified minSup value

        :param value: user specified minSup value

        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value

    def _frequentOneItem(self):
        """
        Generating One frequent items sets

        """
        global _lno, _wf, _weights
        self._mapSupport = {}
        _owf = {}
        for tr in self._Database:
            for i in range(1, len(tr)):
                if tr[i] not in self._mapSupport:
                    self._mapSupport[tr[i]] = [int(tr[0]), int(tr[0]), 1]
                else:
                    self._mapSupport[tr[i]][0] = max(self._mapSupport[tr[i]][0], (int(tr[0]) - self._mapSupport[tr[i]][1]))
                    self._mapSupport[tr[i]][1] = int(tr[0])
                    self._mapSupport[tr[i]][2] += 1
        for key in self._mapSupport:
            self._mapSupport[key][0] = max(self._mapSupport[key][0], abs(len(self._Database) - self._mapSupport[key][1]))
        _lno = len(self._Database)
        self._mapSupport = {k: [v[2], v[0]] for k, v in self._mapSupport.items() if v[0] <= self._regularity}
        gmax = max([self._weight[values] for values in self._mapSupport])
        for x, y in self._mapSupport.items():
            _owf[x] = y[0] * gmax
        self._mapSupport = {k: v for k, v in self._mapSupport.items() if v[0] * _owf[k] >= self._WS}
        for x, y in self._mapSupport.items():
            temp = self._weight[x] * y[0]
            _wf[x] = temp
            self._mapSupport[x].append(temp)
        genList = [k for k, v in sorted(self._mapSupport.items(), key=lambda x: x[1], reverse= True)]
        self._rank = dict([(index, item) for (item, index) in enumerate(genList)])
        for x, y in self._rank.items():
            _weights[y] = self._weight[x]
        return genList

    def _updateTransactions(self, itemSet):
        """
        Updates the items in transactions with rank of items according to their support

        :Example: oneLength = {'a':7, 'b': 5, 'c':'4', 'd':3}
                    rank = {'a':0, 'b':1, 'c':2, 'd':3}

        Parameters
        ----------
        itemSet: list of one-frequent items

        -------

        """
        list1 = []
        for tr in self._Database:
            list2 = [int(tr[0])]
            for i in range(1, len(tr)):
                if tr[i] in itemSet:
                    list2.append(self._rank[tr[i]])
            if len(list2) >= 2:
                basket = list2[1:]
                basket.sort()
                list2[1:] = basket[0:]
                list1.append(list2)
        return list1

    @staticmethod
    def _buildTree(transactions, info):
        """
        Builds the tree with updated transactions
        Parameters:
        ----------
            transactions: updated transactions
            info: support details of each item in transactions

        Returns:
        -------
            transactions compressed in fp-tree

        """
        rootNode = _Tree()
        rootNode.info = info.copy()
        for i in range(len(transactions)):
            set1 = [transactions[i][0]]
            rootNode.addTransaction(transactions[i][1:], set1)
        return rootNode

    def _savePeriodic(self, itemSet):
        """
        The duplication items and their ranks
        Parameters:
        ----------
            itemSet: frequent itemSet that generated

        Returns:
        -------
            patterns with original item names.

        """
        temp = str()
        for i in itemSet:
            temp = temp + self._rankDup[i] + " "
        return temp

    def startMine(self):
        """
            main program to start the operation

        """
        global _WS, _regularity, _weights
        self._startTime = _fp._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self._WS is None:
            raise Exception("Please enter the Minimum Support")
        self._creatingItemSets()
        self._WS = self._convert(self._WS)
        self._regularity = self._convert(self._regularity)
        _WS, _regularity, _weights = self._WS, self._regularity, self._weight
        itemSet = self._frequentOneItem()
        updatedTransactions = self._updateTransactions(itemSet)
        for x, y in self._rank.items():
            self._rankDup[y] = x
        info = {self._rank[k]: v for k, v in self._mapSupport.items()}
        _Tree = self._buildTree(updatedTransactions, info)
        patterns = _Tree.generatePatterns([])
        self._finalPatterns = {}
        for k in patterns:
            s = self._savePeriodic(k[0])
            self._finalPatterns[str(s)] = k[1]
        print("Frequent patterns were generated successfully using frequentPatternGrowth algorithm")
        self._endTime = _fp._time.time()
        self._memoryUSS = float()
        self._memoryRSS = float()
        process = _fp._psutil.Process(_fp._os.getpid())
        self._memoryRSS = float()
        self._memoryUSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process

        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process

        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process

        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe

        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b])
            dataframe = _fp._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataframe

    def savePatterns(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file

        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns

        :rtype: dict
        """
        return self._finalPatterns


if __name__ == "__main__":
    _ap = str()
    if len(_fp._sys.argv) == 6 or len(_fp._sys.argv) == 7:
        if len(_fp._sys.argv) == 7:
            _ap = WFRIMiner(_fp._sys.argv[1], _fp._sys.argv[3], _fp._sys.argv[4], _fp._sys.argv[5], _fp._sys.argv[6])
        if len(_fp._sys.argv) == 5:
            _ap = WFRIMiner(_fp._sys.argv[1], _fp._sys.argv[3], _fp._sys.argv[4], _fp._sys.argv[5])
        _ap.startMine()
        _Patterns = _ap.getPatterns()
        print("Total number of Frequent Patterns:", len(_Patterns))
        _ap.savePatterns(_fp._sys.argv[2])
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
    else:
        _ap = WFRIMiner('/Users/Likhitha/Downloads/wsSample.txt', '/Users/Likhitha/Downloads/ws_weight.txt', 2, 4, ' ')
        _ap.startMine()
        _Patterns = _ap.getPatterns()
        for x, y in _Patterns.items():
            print(x, y)
        print("Total number of Patterns:", len(_Patterns))
        _ap.savePatterns('/Users/Likhitha/Downloads/output.txt')
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
        print("Error! The number of input parameters do not match the total number of parameters provided")



from pyspark.context import SparkContext
from pyspark.conf import SparkConf
# import abstract as _ab
from PAMI.frequentPattern.pyspark import abstract as _ab
from collections import OrderedDict

class Node:
    """
    Attribute
    ---------
        item : int
            Storing item of a node
        count : int
            To maintain the support count of node
        children : dict
            To maintain the children of node
        prefix : list
            To maintain the prefix of node
    """

    def __init__(self,item,count,children):
        self.item = item
        self.count = count
        self.children = children
        self.prefix = []


class Tree:
    """
    Attribute
    ---------
        root : Node
            The first node of the tree set to Null
        nodeLink : dict
            Store nodes that have the same item

    Methods
    -------
        createTree(transaction,count)
            Create tree from transaction and count
        linkNode(node)
            Add nodes that have the same item to self.nodeLink
        createCPB(item)
            Create conditional pattern base of item
        mergeTree(tree,fpList)
            Merge tree
        createTransactions(fpList)
            Create transactions from yourself(tree)
        getPattern(item,suffixItem,minSup)
            Get frequent patterns about suffixItem from yourself
        mining(minSup,isResponsible = lambda x:True)
            Do pattern mining on your own
    """


    def __init__(self):
        self.root = Node(None,0,{})
        self.nodeLink = OrderedDict()


    def createTree(self,transaction,count):
        """
        Adding transaction into tree
        :param transaction: list
        :param count: int
        :return: Tree
        """
        current = self.root
        for item in transaction:
            if item not in current.children:
                current.children[item] = Node(item,count,{})
                current.children[item].prefix = transaction[0:transaction.index(item)]
                self.linkNode(current.children[item])
            else:
                current.children[item].count += count
            current = current.children[item]
        return self


    def linkNode(self, node):
        """
        Adding node to nodeLinks
        :param node: Node
        :return:
        """
        if node.item in self.nodeLink:
            self.nodeLink[node.item].append(node)
        else:
            self.nodeLink[node.item] = []
            self.nodeLink[node.item].append(node)


    def createCPB(self,item):
        """
        Create conditional pattern base based on item
        :param item: int
        :return: Tree
        """
        pTree = Tree()
        for node in self.nodeLink[item]:
            pTree.createTree(node.prefix,node.count)
        return pTree


    def mergeTree(self,tree,fpList):
        """
        Merging tree to self
        :param tree: Tree
        :param fpList: list
        :return: Tree
        """
        transactions = tree.createTransactions(fpList)
        for transaction in transactions:
            self.createTree(transaction,1)
        return self


    def createTransactions(self,fpList):
        """
        Creating transactions from self
        :param fpList: list
        :return: list
        """
        transactions = []
        flist = [x for x in fpList if x in self.nodeLink]
        for item in reversed(flist):
            for node in self.nodeLink[item]:
                if node.count != 0:
                    transaction = node.prefix
                    transaction.append(node.item)
                    transactions.extend([transaction for i in range(node.count)])
                    current = self.root
                    for i in transaction:
                        current = current.children[i]
                        current.count -= node.count
        return transactions


    def getPattern(self,item,suffixItem,minSup):
        """
        Getting frequent patterns about suffixItem
        :param item: int
        :param suffixItem: tuple
        :param minSup: int
        :return: list
        """
        pTree = self.createCPB(item)
        frequentItems = {}
        frequentPatterns = []
        for i in pTree.nodeLink.keys():
            frequentItems[i] = 0
            for node in pTree.nodeLink[i]:
                frequentItems[i] += node.count
        frequentItems = {key: value for key, value in frequentItems.items() if value >= minSup}
        for i in frequentItems:
            pattern = list(suffixItem)
            pattern.append(i)
            frequentPatterns.append((tuple(pattern),frequentItems[i]))
            frequentPatterns.extend(pTree.getPattern(i, tuple(pattern), minSup))
        return frequentPatterns


    def mining(self,minSup,isResponsible = lambda x:True):
        """
        Mining self and return frequent patterns
        :param minSup: int
        :param isResponsible: function
        :return: list
        """
        frequentPatterns = []
        flist = sorted([item for item in self.nodeLink.keys()])
        for item in reversed(flist):
            if isResponsible(item):
                frequentPatterns.extend(self.getPattern(item,tuple([item]),minSup))
        return frequentPatterns



class parallelFPGrowth(_ab._frequentPatterns):
    """
        Attributes
        ----------
            minSup : float
                The user can specify minSup either in count or proportion of database size.
            iFile : file
                Input file name or path of the input file.
            oFile : file
                Name of the output file or the path of the output file.
            sep : str
                This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
                However, the users can override their default separator.
            startTime:float
                To record the start time of the mining process
            endTime:float
                To record the completion time of the mining process
            memoryUSS : float
                To store the total amou.nt of USS memory consumed by the program
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program
            finalPatterns : dict
                it represents to store the all frequent patterns
            numWorkers: int
                The number of workers
                This value means the number of cores which are used.

        Methods
        -------
            startMine()
                Mining process will start from this function
            getPatterns()
                Complete set of patterns will be retrieved with this function
            savePatterns(outFile)
                Complete set of frequent patterns will be loaded in to a output file
            getPatternsAsDataFrame()
                Complete set of frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function
            getFrequentItems(data)
                Get frequent items that length is 1.
            genCondTransaction(data, rank)
                Generating conditional transactions for distributed pattern mining
            getPartitionId(item)
                Get partition id of item
                FPTree is created on each workers based on partition id.
            getAllFrequentPatterns(data, fpList)
                Get all frequent patterns
    """

    _minSup = float()
    _numWorkers = int()
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()

    def __init__(self, iFile, minSup, numWorkers, sep = '\t'):
        super().__init__(iFile, float(minSup), int(numWorkers), sep)

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function
        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function
        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process
        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe
        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def savePatterns(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to a output file
        :param outFile: name of the output file
        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            if type(x) == tuple:
                pattern = ""
                for item in x:
                    pattern = pattern + str(item) + " "
                s1 = pattern + ":" + str(y)
            else:
                s1 = str(x) + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """
        Function to send the set of frequent patterns after completion of the mining process
        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def _getFrequentItems(self, data):
        """
        Get the frequent items that length is 1
        :param data: RDD
        :return: dict
        """
        flatData = data.flatMap(lambda x: [(y, 1) for y in x])
        oneFrequentItems = dict(flatData.reduceByKey(lambda x, y: x + y)
                                .filter(lambda c: c[1] >= self._minSup).collect())
        return oneFrequentItems

    def _getAllFrequentPatterns(self, data, FPList):
        """
        Get all frequent patterns
        :param data: RDD
        :param fpList: list
        :return: list
        """
        rank = dict([(index, item) for (item, index) in enumerate(FPList)])
        newFPList = list(rank.values())
        workByPartition = data.flatMap(lambda transaction: self._genCondTransaction(transaction, rank))
        emptyTree = Tree()
        forest = workByPartition.aggregateByKey(emptyTree, lambda tree, transaction: tree.createTree(transaction, 1),
                                                lambda tree1, tree2: tree1.mergeTree(tree2, newFPList))
        frequentItemsets = forest.flatMap(lambda tree_tuple:
                                          tree_tuple[1].mining(self._minSup,
                                                               lambda x: self._getPartitionId(x) == tree_tuple[0]))
        I = frequentItemsets.map(lambda ranks_count: ([FPList[int(z)] for z in ranks_count[0]], ranks_count[1]))
        return I

    def _getPartitionId(self, item):
        """
        Get partition id of item
        :param item: int
        :return: int
        """
        return item % self._numWorkers

    def _genCondTransaction(self, transaction, rank):
        """
        Generate conditional transactions from transaction
        :param transaction : list
        :param rank: dict
        :return: list
        """
        filtered = [rank[int(x)] for x in transaction if int(x) in rank.keys()]
        filtered = sorted(filtered)
        condTransaction = {}
        for i in range(len(filtered) - 1, -1, -1):
            item = filtered[i]
            partition = self._getPartitionId(item)
            if partition not in condTransaction:
                condTransaction[partition] = filtered[:i + 1]
        return [x for x in condTransaction.items()]

    def startMine(self):
        """
        Main program to start the operation
        """
        self._startTime = _ab._time.time()
        conf = SparkConf().setAppName("FPGrowth").setMaster("local[*]")
        sc = SparkContext(conf=conf)
        #sc.addFile("fpTree.py")

        data = sc.textFile(self._iFile, self._numWorkers).map(lambda x: [int(y) for y in x.strip().split(self._sep)])\
            .persist()

        frequentItems = self._getFrequentItems(data)
        self._finalPatterns.update(frequentItems)

        FPList = [x for (x, y) in sorted(frequentItems.items(), key=lambda x: -x[1])]
        frequentPatterns = self._getAllFrequentPatterns(data, FPList)

        for v in list(frequentPatterns.collect()):
            temp = (tuple(v[0]), v[1])
            self._finalPatterns[temp[0]] = temp[1]

        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        sc.stop()




if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:
            _ap = PFPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:
            _ap = PFPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        _ap.startMine()
        _finalPatterns = _ap.getPatterns()
        print("Total number of Frequent Patterns:", len(_finalPatterns))
        _ap.savePatterns(_ab._sys.argv[2])
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



import os
import csv
import time
import numpy as np
import pycuda.gpuarray as gpuarray
import pycuda.autoinit
import psutil
import pycuda.driver as cuda
from pycuda.compiler import SourceModule
import pycuda

deviceIntersection = SourceModule("""
    __global__ void intersection(int *compareThis, int *compareThat, int *resultStart,
                                 int *values, int *result, int resultX, int resultY){
        const int tidX = blockIdx.x * blockDim.x + threadIdx.x;
        const int tidY = blockIdx.y * blockDim.y + threadIdx.y; 
        int resultIndex = resultStart[tidX] + tidY;
        
        // ignore if tidX or tidY is out of bounds or if the value comparing with is 0
        if (tidX > resultX-1 || tidY > resultY-1 || values[compareThis[tidX] + tidY] == 0) return;

        for (int i = 0; i < resultY; i++){
            if ( values[compareThat[tidX] + i] == 0) return;
            if ( values[compareThis[tidX] + tidY] == values[compareThat[tidX] + i]){
                result[resultIndex] = values[compareThis[tidX] + tidY];
                return;
            }
        }

        //result[resultIndex] = values[compareThis[tidX] + tidY];

    }

"""
                                  )


class cudaAprioriTID:
    __time = 0
    __memRSS = 0
    __memUSS = 0
    __GPU_MEM = 0
    filePath = ""
    sep = ""
    minSup = 0
    Patterns = {}

    def __init__(self, filePath, sep, minSup):
        self.filePath = filePath
        self.sep = sep
        self.minSup = minSup
        self.__time = 0
        self.__memRSS = 0
        self.__memUSS = 0

    def _readFile(self, fileName, separator):
        """
        Reads a file and stores the data in a dictionary

        Args:
            fileName: string
            separator: string

        Returns:
            dictionary: dictionary
        """
        file = open(fileName, 'r')
        dictionary = {}
        lineNumber = 1
        for line in file:
            line = line.strip()
            line = line.split(separator)
            for i in range(len(line)):
                if line[i] in dictionary:
                    dictionary[line[i]].append(lineNumber)
                else:
                    dictionary[line[i]] = [lineNumber]
            lineNumber += 1

        # sort dictionary by size of values
        dictionary = dict(
            sorted(dictionary.items(), key=lambda x: len(x[1]), reverse=True))
        return dictionary, lineNumber

    def get_time(self):
        return self.__time

    def get_memRSS(self):
        return self.__memRSS

    def get_memUSS(self):
        return self.__memUSS

    def get_GPU_MEM(self):
        return self.__GPU_MEM

    def get_Patterns(self):
        return self.Patterns

    def get_numberOfPatterns(self):
        return len(self.Patterns)

    def startMine(self):
        dev_Intersection = deviceIntersection.get_function("intersection")
        startTime = time.time()
        final = {}

        data, lineNo = self._readFile(self.filePath, self.sep)
        if self.minSup < 1:
            self.minSup = int(lineNo * self.minSup)

        data = dict(filter(lambda x: len(x[1]) >= self.minSup, data.items()))
        for key, value in data.items():
            final[key] = len(value)

        while len(data) > 1:
            # sort data by size of values
            data = dict(
                sorted(data.items(), key=lambda x: len(x[1]), reverse=True))

            values = list(data.values())
            maxLength = values[0]
            for i in range(1, len(values)):
                while len(values[i]) != len(maxLength):
                    values[i].append(0)

            values = np.array(values)
            resultSize = 0

            compareThis = []
            compareThat = []
            resultStart = []
            counter = 0

            for i in range(len(values)):
                for j in range(i+1, len(values)):
                    resultSize += 1
                    compareThis.append(i*len(maxLength))
                    compareThat.append(j*len(maxLength))
                    resultStart.append(counter)
                    counter += len(maxLength)
            result = np.zeros((resultSize, len(maxLength)), dtype=np.int32)

            # convert all to uint32
            compareThis = np.array(compareThis, dtype=np.uint32)
            compareThat = np.array(compareThat, dtype=np.uint32)
            resultStart = np.array(resultStart, dtype=np.uint32)
            values = np.array(values, dtype=np.uint32)
            result = np.array(result, dtype=np.uint32)

            # allocate memory on GPU
            compareThis_gpu = cuda.mem_alloc(compareThis.nbytes)
            compareThat_gpu = cuda.mem_alloc(compareThat.nbytes)
            resultStart_gpu = cuda.mem_alloc(resultStart.nbytes)
            values_gpu = cuda.mem_alloc(values.nbytes)
            result_gpu = cuda.mem_alloc(result.nbytes)

            # add all nbytes to GPU_MEM
            sumBytes = compareThis.nbytes + compareThat.nbytes + resultStart.nbytes + values.nbytes + result.nbytes
            if sumBytes > self.__GPU_MEM:
                self.__GPU_MEM = sumBytes

            # copy data to GPU
            cuda.memcpy_htod(compareThis_gpu, compareThis)
            cuda.memcpy_htod(compareThat_gpu, compareThat)
            cuda.memcpy_htod(resultStart_gpu, resultStart)
            cuda.memcpy_htod(values_gpu, values)
            cuda.memcpy_htod(result_gpu, result)

            blockDim = (32, 32, 1)
            gridDim = (resultSize//32 + 1, len(maxLength)//32 + 1, 1)

            dev_Intersection(compareThis_gpu, compareThat_gpu,
                             resultStart_gpu, values_gpu, result_gpu,
                             np.uint32(resultSize), np.uint32(len(maxLength)),
                             block=blockDim, grid=gridDim)

            # copy data back to CPU
            cuda.Context.synchronize()
            cuda.memcpy_dtoh(result, result_gpu)

            # free GPU memory
            cuda.DeviceAllocation.free(compareThis_gpu)
            cuda.DeviceAllocation.free(compareThat_gpu)
            cuda.DeviceAllocation.free(resultStart_gpu)
            cuda.DeviceAllocation.free(values_gpu)
            cuda.DeviceAllocation.free(result_gpu)

            keys = list(data.keys())
            # convert all to string and add " "
            for i in range(len(keys)):
                keys[i] = str(keys[i]) + " "
            data = {}
            index = 0
            for i in range(len(keys)):
                for j in range(i+1, len(keys)):
                    newResult = list(sorted(set(result[index])))
                    newResult = list(filter(lambda x: x > 0, newResult))
                    if len(newResult) >= self.minSup:
                        keyI = keys[i].split()
                        keyJ = keys[j].split()
                        combinedKey = " ".join(list(str(x) for x in (
                            sorted(int(x) for x in (set(keyI) | set(keyJ))))))
                        if combinedKey not in final:
                            data[combinedKey] = newResult
                            final[combinedKey] = len(newResult)
                    index += 1


        self.__time = time.time() - startTime
        self.__memRSS = psutil.Process(os.getpid()).memory_info().rss
        self.__memUSS = psutil.Process(os.getpid()).memory_full_info().uss
        self.Patterns = final


if __name__ == "__main__":
    filePath = "datasets\\transactional_T10I4D100K.csv"
    sep = "\t"
    support = 500
    cudaAprioriTID = cudaAprioriTID(filePath, sep, support)
    cudaAprioriTID.startMine()
    print("Time: ", cudaAprioriTID.get_time())
    print("Memory RSS: ", cudaAprioriTID.get_memRSS())
    print("Memory USS: ", cudaAprioriTID.get_memUSS())
    print("GPU MEM: ", cudaAprioriTID.get_GPU_MEM())
    print("Patterns: ", cudaAprioriTID.get_numberOfPatterns())



#  Copyright (C)  2021 Rage Uday Kiran
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from PAMI.periodicFrequentPattern.basic import abstract as _ab


class PFECLAT(_ab._periodicFrequentPatterns):
    """ EclatPFP is the fundamental approach to mine the periodic-frequent patterns.

        Reference:
        --------
            P. Ravikumar, P.Likhitha, R. Uday kiran, Y. Watanobe, and Koji Zettsu, "Towards efficient discovery of 
            periodic-frequent patterns in columnar temporal databases", 2021 IEA/AIE.

    Attributes:
    ----------
        iFile : file
            Name of the Input file or path of the input file
        oFile : file
            Name of the output file or path of the output file
        minSup: int or float or str
            The user can specify minSup either in count or proportion of database size.
            If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
            Otherwise, it will be treated as float.
            Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
        maxPer: int or float or str
            The user can specify maxPer either in count or proportion of database size.
            If the program detects the data type of maxPer is integer, then it treats maxPer is expressed in count.
            Otherwise, it will be treated as float.
            Example: maxPer=10 will be treated as integer, while maxPer=10.0 will be treated as float
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
            However, the users can override their default separator.
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        Database : list
            To store the transactions of a database in list
        mapSupport : Dictionary
            To maintain the information of item and their frequency
        lno : int
            it represents the total no of transactions
        tree : class
            it represents the Tree class
        itemSetCount : int
            it represents the total no of patterns
        finalPatterns : dict
            it represents to store the patterns
        tidList : dict
            stores the timestamps of an item
        hashing : dict
            stores the patterns with their support to check for the closed property

    Methods:
    -------
        startMine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        savePatterns(oFile)
            Complete set of periodic-frequent patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of periodic-frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        creatingOneItemSets()
            Scan the database and store the items with their timestamps which are periodic frequent 
        getPeriodAndSupport()
            Calculates the support and period for a list of timestamps.
        Generation()
            Used to implement prefix class equivalence method to generate the periodic patterns recursively
            
        Executing the code on terminal:
        -------
        Format:
        ------
            python3 PFPECLAT.py <inputFile> <outputFile> <minSup>

        Examples:
        --------
            python3 PFPECLAT.py sampleDB.txt patterns.txt 10.0   (minSup will be considered in percentage of database transactions)

            python3 PFPECLAT.py sampleDB.txt patterns.txt 10     (minSup will be considered in support count or frequency)
        
        Sample run of the imported code:
        --------------
        
            from PAMI.periodicFrequentPattern.basic import PFPECLAT as alg

            obj = alg.PFPECLAT("../basic/sampleTDB.txt", "2", "5")

            obj.startMine()

            periodicFrequentPatterns = obj.getPatterns()

            print("Total number of Periodic Frequent Patterns:", len(periodicFrequentPatterns))

            obj.savePatterns("patterns")

            Df = obj.getPatternsAsDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

        Credits:
        -------
            The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.\n

        """
    
    _iFile = " "
    _oFile = " "
    _sep = " "
    _dbSize = None
    _Database = None
    _minSup = str()
    _maxPer = str()
    _tidSet = set()
    _finalPatterns = {}
    _startTime = None
    _endTime = None
    _memoryUSS = float()
    _memoryRSS = float()

    def _getPeriodic(self, tids: set):
        tidList = list(tids)
        tidList.sort()
        tidList.append(self._dbSize)
        cur = 0
        per = 0
        for tid in tidList:
            per = max(per, tid - cur)
            if per > self._maxPer:  # early stopping
                break
            cur = tid
        return per

    def _convert(self, value):
        """
        To convert the given user specified value

        :param value: user specified value
        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self._dbSize * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (self._dbSize * value)
            else:
                value = int(value)
        return value

    def _creatingOneItemSets(self):
        """Storing the complete transactions of the database/input file in a database variable
        """
        plist = []
        Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            ts, data = [], []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'TS' in i:
                ts = self._iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self._iFile['Transactions'].tolist()
            for i in range(len(data)):
                tr = [ts[i][0]]
                tr = tr + data[i]
                Database.append(tr)
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()
        tid = 0
        itemsets = {}  # {key: item, value: list of tids}
        periodicHelper = {}  # {key: item, value: [period, last_tid]}
        for line in Database:
            tid = int(line[0])
            self._tidSet.add(tid)
            for item in line[1:]:
                if item in itemsets:
                    itemsets[item].add(tid)
                    periodicHelper[item][0] = max(periodicHelper[item][0],
                                                  abs(tid - periodicHelper[item][1]))  # update current max period
                    periodicHelper[item][1] = tid  # update the last tid
                else:
                    itemsets[item] = {tid}
                    periodicHelper[item] = [abs(0 - tid), tid]  # initialize helper

        # finish all items' period
        self._dbSize = len(Database)
        self._minSup = self._convert(self._minSup)
        self._maxPer = self._convert(self._maxPer)
        del Database
        for item, _ in periodicHelper.items():
            periodicHelper[item][0] = max(periodicHelper[item][0],
                                          abs(self._dbSize - periodicHelper[item][1]))  # tid of the last transaction
        candidates = []
        for item, tids in itemsets.items():
            per = periodicHelper[item][0]
            sup = len(tids)
            if sup >= self._minSup and per <= self._maxPer:
                candidates.append(item)
                self._finalPatterns[item] = [sup, per, tids]
        return candidates
    
    def _generateEclat(self, candidates):
        newCandidates = []
        for i in range(0, len(candidates)):
            prefixItem = candidates[i]
            prefixItemSet = prefixItem.split()
            for j in range(i + 1, len(candidates)):
                item = candidates[j]
                itemSet = item.split()
                if prefixItemSet[:-1] == itemSet[:-1] and prefixItemSet[-1] != itemSet[-1]:
                    _value = self._finalPatterns[item][2].intersection(self._finalPatterns[prefixItem][2])
                    sup = len(_value)
                    per = self._getPeriodic(_value)
                    if sup >= self._minSup and per <= self._maxPer:
                        newItem = prefixItem + " " + itemSet[-1]
                        self._finalPatterns[newItem] = [sup, per, _value]
                        newCandidates.append(newItem)

        if len(newCandidates) > 0:
            self._generateEclat(newCandidates)
    
    def startMine(self):
        #print(f"Optimized {type(self).__name__}")
        self._startTime = _ab._time.time()
        self._finalPatterns = {}
        frequentSets = self._creatingOneItemSets()
        self._generateEclat(frequentSets)
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryRSS = float()
        self._memoryUSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Periodic-Frequent patterns were generated successfully using PFECLAT algorithm ")

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final periodic-frequent patterns in a dataframe

        :return: returning periodic-frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b[0], b[1]])
            dataframe = _ab._pd.DataFrame(data, columns=['Patterns', 'Support', 'Periodicity'])
        return dataframe

    def savePatterns(self, outFile):
        """Complete set of periodic-frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x + ":" + str(y[0]) + ":" + str(y[1])
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of periodic-frequent patterns after completion of the mining process

        :return: returning periodic-frequent patterns
        :rtype: dict
        """
        return self._finalPatterns
                    

if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:
            _ap = PFECLAT(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:
            _ap = PFECLAT(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        _ap.startMine()
        _Patterns = _ap.getPatterns()
        print("Total number of Periodic-Frequent Patterns:", len(_Patterns))
        _ap.savePatterns(_ab._sys.argv[2])
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
    else:
        '''_ap = PFECLAT('/Users/Likhitha/Downloads/dense_DB.csv', 23, 25, ',')
        _ap.startMine()
        _Patterns = _ap.getPatterns()
        print("Total number of Patterns:", len(_Patterns))
        _ap.savePatterns('/Users/Likhitha/Downloads/output.txt')
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)'''
        print("Error! The number of input parameters do not match the total number of parameters provided")



import sys
import time
import resource
import math
import itertools
path=sys.argv[1]
output = sys.argv[2]
minSup = int(sys.argv[3])
itemSup = int(sys.argv[4])
minLength = int(sys.argv[5])
faultTolerance = int(sys.argv[6])
mapSupport = {}
finalPatterns = {}
Database = []

def Count(k):
    count = 0
    items = []
    k = list(k)
    n = len(k)-faultTolerance
    c = itertools.combinations(k, n)
    count = 0
    for j in c: 
        j = list(j)
        for i in Database:
            if set(j).issubset(i):
                count += 1
                items.append(i)
    items =  list(set(map(tuple,items)))
    print(k, items, count)
    return len(items), items
            
    
def countItemSupport(itemset, transactions):
    tids = {}
    res = True
    for i in itemset:
        for k in transactions:
            if i in k:
                if i not in tids:
                    tids[i] = 1
                else:
                    tids[i] += 1
    #print(itemset, transactions, tids)
    for x, y in tids.items():
        if y < itemSup:
            res = False
    return res

def getFaultPatterns():
    
    l = [k for k,v in mapSupport.items()]
    for i in range(2, len(l)+1):
        c = itertools.combinations(l,i)
        for j in c:
            support, items = Count(j)
            res = countItemSupport(j, items)
            if len(j)> minLength and len(j)>=faultTolerance and support >= minSup and res == True:
                finalPatterns[tuple(j)] = support
                    

with open(path, 'r') as f:
    for line in f:
        li = line.split()
        Database.append(li)
        for i in li:
            if i not in mapSupport:
                mapSupport[i] = 1
            else:
                mapSupport[i] +=1
l = [k for k, v in mapSupport.items()]
for i in range(len(l)):
    for j in range(i+1, len(l)):
        x, y = l[i], l[j]
        li = [x, y]
        count = 0
        tids = {x:0, y:0}
        for k in Database:
            if x in k and y in k:
                count += 1
                tids[x] += 1
                tids[y] += 1
            if x in k and y not in k:
                count += 1
                tids[x] = 1
            if x not in k and y in k:
                count += 1
                tids[y] += 1
        re = True
        for x, y in tids.items():
            if y< itemSup:
                re = False
        finalPatterns[tuple(li)] = count

getFaultPatterns() 

print("...")
with open(output, 'w+') as f:
    for x,y in finalPatterns.items():
        s = str(x) + ":" + str(y)
        f.write("%s \n" % s)
for x,y in finalPatterns.items():
    print(x,y)



import pandas as pd

class denseDF2DB:
    """
        This class create Data Base from DataFrame.

        Attribute:
        ----------
        inputDF : pandas.DataFrame
            It is dense DataFrame
        condition : str
            It is condition to judge the value in dataframe
        thresholdValue : int or float:
            User defined value.
        tids : list
            It is tids list.
        items : list
            Store the items list
        outputFile : str
            Creation data base output to this outputFile.

        Methods:
        --------
        createDB(outputFile)
            Create transactional data base from dataFrame
        createTDB(outputFile)
            Create temporal dataBase from dataFrame
        createUDB(outputFile)
            Create utility database from dataFrame
        getFileName()
            Return outputFileName.
        """

    def __init__(self, inputDF, condition, thresholdValue):
        self.inputDF = inputDF
        self.condition = condition
        self.thresholdValue = thresholdValue
        self.tids = []
        self.items = []
        self.outputFile = ' '
        self.items = list(self.inputDF.columns.values)[1:]
        self.inputDF = self.inputDF.set_index('tid')
        self.tids = list(self.inputDF.index)


    def createTransactional(self, outputFile):
        """
        Create transactional data base

        :param outputFile: Write transactional data base into outputFile
        :type outputFile: str
        """

        self.outputFile = outputFile
        with open(outputFile, 'w') as f:
            if self.condition == '>':
                for tid in self.tids:
                    transaction = [item for item in self.items if self.inputDF.at[tid, item] > self.thresholdValue]
                    if len(transaction) > 1:
                        f.write(f'{transaction[0]}')
                        for item in transaction[1:]:
                            f.write(f',{item}')
                    elif len(transaction) == 1:
                        f.write(f'{transaction}')
                    else:
                        continue
                    f.write('\n')

            elif self.condition == '>=':
                for tid in self.tids:
                    transaction = [item for item in self.items if self.inputDF.at[tid, item] >= self.thresholdValue]
                    if len(transaction) > 1:
                        f.write(f'{transaction[0]}')
                        for item in transaction[1:]:
                            f.write(f',{item}')
                    elif len(transaction) == 1:
                        f.write(f'{transaction}')
                    else:
                        continue
                    f.write('\n')

            elif self.condition == '<=':
                for tid in self.tids:
                    transaction = [item for item in self.items if self.inputDF.at[tid, item] <= self.thresholdValue]
                    if len(transaction) > 1:
                        f.write(f'{transaction[0]}')
                        for item in transaction[1:]:
                            f.write(f',{item}')
                    elif len(transaction) == 1:
                        f.write(f'{transaction}')
                    else:
                        continue
                    f.write('\n')

            elif self.condition == '<':
                for tid in self.tids:
                    transaction = [item for item in self.items if self.inputDF.at[tid, item] < self.thresholdValue]
                    if len(transaction) > 1:
                        f.write(f'{transaction[0]}')
                        for item in transaction[1:]:
                            f.write(f',{item}')
                    elif len(transaction) == 1:
                        f.write(f'{transaction}')
                    else:
                        continue
                    f.write('\n')
            elif self.condition == '==':
                for tid in self.tids:
                    transaction = [item for item in self.items if self.inputDF.at[tid, item] == self.thresholdValue]
                    if len(transaction) > 1:
                        f.write(f'{transaction[0]}')
                        for item in transaction[1:]:
                            f.write(f',{item}')
                    elif len(transaction) == 1:
                        f.write(f'{transaction}')
                    else:
                        continue
                    f.write('\n')
            elif self.condition == '!=':
                for tid in self.tids:
                    transaction = [item for item in self.items if self.inputDF.at[tid, item] != self.thresholdValue]
                    if len(transaction) > 1:
                        f.write(f'{transaction[0]}')
                        for item in transaction[1:]:
                            f.write(f',{item}')
                    elif len(transaction) == 1:
                        f.write(f'{transaction}')
                    else:
                        continue
                    f.write('\n')
            else:
                print('Condition error')



    def createTemporal(self, outputFile):
        """
        Create temporal data base

        :param outputFile: Write temporal data base into outputFile
        :type outputFile: str
        """

        self.outputFile = outputFile
        with open(outputFile, 'w') as f:
            if self.condition == '>':
                for tid in self.tids:
                    transaction = [item for item in self.items if self.inputDF.at[tid, item] > self.thresholdValue]
                    if len(transaction) > 1:
                        f.write(f'{tid}')
                        for item in transaction:
                            f.write(f',{item}')
                    elif len(transaction) == 1:
                        f.write(f'{tid}')
                        f.write(f',{transaction}')
                    else:
                        continue
                    f.write('\n')

            elif self.condition == '>=':
                for tid in self.tids:
                    transaction = [item for item in self.items if self.inputDF.at[tid, item] >= self.thresholdValue]
                    if len(transaction) > 1:
                        f.write(f'{tid}')
                        for item in transaction:
                            f.write(f',{item}')
                    elif len(transaction) == 1:
                        f.write(f'{tid}')
                        f.write(f',{transaction}')
                    else:
                        continue
                    f.write('\n')

            elif self.condition == '<=':
                for tid in self.tids:
                    transaction = [item for item in self.items if self.inputDF.at[tid, item] <= self.thresholdValue]
                    if len(transaction) > 1:
                        f.write(f'{tid}')
                        for item in transaction:
                            f.write(f',{item}')
                    elif len(transaction) == 1:
                        f.write(f'{tid}')
                        f.write(f',{transaction}')
                    else:
                        continue
                    f.write('\n')

            elif self.condition == '<':
                for tid in self.tids:
                    transaction = [item for item in self.items if self.inputDF.at[tid, item] < self.thresholdValue]
                    if len(transaction) > 1:
                        f.write(f'{tid}')
                        for item in transaction:
                            f.write(f',{item}')
                    elif len(transaction) == 1:
                        f.write(f'{tid}')
                        f.write(f',{transaction}')
                    else:
                        continue
                    f.write('\n')
            elif self.condition == '==':
                for tid in self.tids:
                    transaction = [item for item in self.items if self.inputDF.at[tid, item] == self.thresholdValue]
                    if len(transaction) > 1:
                        f.write(f'{tid}')
                        for item in transaction:
                            f.write(f',{item}')
                    elif len(transaction) == 1:
                        f.write(f'{tid}')
                        f.write(f',{transaction}')
                    else:
                        continue
                    f.write('\n')
            elif self.condition == '!=':
                for tid in self.tids:
                    transaction = [item for item in self.items if self.inputDF.at[tid, item] != self.thresholdValue]
                    if len(transaction) > 1:
                        f.write(f'{tid}')
                        for item in transaction:
                            f.write(f',{item}')
                    elif len(transaction) == 1:
                        f.write(f'{tid}')
                        f.write(f',{transaction}')
                    else:
                        continue
                    f.write('\n')

            else:
                print('Condition error')

    def createUtility(self, outputFile):
        """
        Create the utility data base.

        :param outputFile: Write utility data base into outputFile
        :type outputFile: str
        """

        self.outputFile = outputFile
        with open(self.outputFile, 'w') as f:
            for tid in self.tids:
                df = self.inputDF.loc[tid].dropna()
                f.write(f'{df.index[0]}')
                for item in df.index[1:]:
                    f.write(f'\t{item}')
                f.write(f':{df.sum()}:')
                f.write(f'{df.at[df.index[0]]}')
                for item in df.index[1:]:
                    f.write(f'\t{df.at[item]}')
                f.write('\n')

    def getFileName(self):
        """
        return outputFile name

        :return: outputFile name
        """

        return self.outputFile


if __name__ == '__main__':
    DF = createDenseDF('denseDF.csv')
    obj = denseDF2DB(DF.getDF(), '>=', 2)
    obj.createDB('testTransactional.csv')
    transactionalDB = obj.getFileName()
    print(transactionalDB)



import random


class generateTransactionalDatabase:
    """
    generateTransactionalDatabase generates a transactional database

        Attributes:
        -----------
        numOfTransactions: int
            number of transactions
        maxNumOfDistinctItems: int
            maximum number of distinct items
        numOfItemsPerTransaction: int
            number of items per transaction
        outFileName: str
            output file name
        sep: str
            seperator in file, default is tab space

        Methods:
        --------
        getFileName()
            get output filename
    """
    def __init__(self, numOfTransactions, maxNumOfDistinctItems, numOfItemsPerTransaction, outFileName, sep='\t'):
        """

        :param numOfTransactions: number of transactions
        :type numOfTransactions: int
        :param maxNumOfDistinctItems: distinct items per transactions
        :type maxNumOfDistinctItems: int
        :param numOfItemsPerTransaction: items per transaction
        :type numOfItemsPerTransaction: int
        :param outFileName: output filename
        :type outFileName: str
        :param sep: seperator
        :type sep: str
        """
        self.numOfTransactions = numOfTransactions
        self.maxNumOfDistinctItems = maxNumOfDistinctItems
        self.numOfItemsPerTransaction = numOfItemsPerTransaction
        self.outFileName = outFileName
        self.sep = sep

        # make outFile
        with open(self.outFileName, "w+") as outFile:
            # For the number of transactions to be generated
            for i in range(self.numOfTransactions):
                # This hashset will be used to remember which items have
                # already been added to this item set.
                alreadyAdded = set()
                # create an arraylist to store items from the item set that will be generated
                itemSet = list()
                # We randomly decide how many items will appear in this transaction
                randNumOfItems = random.randrange(self.maxNumOfDistinctItems) + 1
                # for the number of items that was decided above
                for j in range(randNumOfItems):
                    # we generate the item randomly and write it to disk
                    item = random.randrange(self.maxNumOfDistinctItems) + 1
                    # if we already added this item to this item set
                    # we choose another one
                    while item in alreadyAdded:
                        item = random.randrange(self.maxNumOfDistinctItems) + 1
                    alreadyAdded.add(item)
                    itemSet.append(item)
                # sort the item set
                itemSet.sort()
                # write the item set
                for j in itemSet:
                    outFile.write(str(j) + self.sep)
                outFile.write('\n')
        # close outFile
        outFile.close()



    def getFileName(self):
        """
        return output file name
        :return: output file name
        """
        return self.outFileName

if __name__ == '__main__':
    numOfTransactions = 500
    maxNumOfDistinctItems = 1000
    numOfItemsPerTransaction = 20
    outFileName = '/Users/Likhitha/Downloads/out.txt'

    tDG = generateTransactionalDatabase(numOfTransactions, maxNumOfDistinctItems, numOfItemsPerTransaction, outFileName)



from pyspark import SparkConf, SparkContext
# import abstract as _ab
from PAMI.frequentPattern.pyspark import abstract as _ab


class parallelECLAT(_ab._frequentPatterns):
    """
        Attributes
        ----------
            iFile : file
                Input file name or path of the input file
            oFile : file
                Name of the output file or the path of output file
            minSup : int
                minSup is a minimum support
            numWorkers : int
                the number of workers
            finalPatterns : dict
                Storing the complete set of patterns in a dictionary variable
            startTime:float
                To record the start time of the mining process
            endTime:float
                To record the completion time of the mining process
            memoryUSS : float
                To store the total amount of USS memory consumed by the program
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program

        Methods:
        -------
            startMine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            storePatternsInFile(oFile)
                Complete set of frequent patterns will be loaded in to a output file
            getPatternsInDataFrame()
                Complete set of frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function
            genPatterns(suffix, pattern, data)
                Generate frequent pattern from the arguments

        Executing the code on terminal:
        -------------------------------

            Format:
            ------
                python3 parallelApriori.py <inputFile> <outputFile> <minSup> <numWorkers>

            Examples:
            ---------
                python3 Apriori.py sampleDB.txt patterns.txt 1000 10  (minSup will be considered in support count or frequency)
        """

    _minSup = float()
    _numWorkers = int()
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()

    def __init__(self, iFile, minSup, numWorkers, sep='\t'):
        super().__init__(iFile, float(minSup), int(numWorkers), sep)

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function
        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function
        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process
        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe
        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def savePatterns(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to a output file
        :param outFile: name of the output file
        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """
        Function to send the set of frequent patterns after completion of the mining process
        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def _genPatterns(self, suffix, pattern, data):
        freqPatterns = {}
        index = data.index(suffix)
        for i in range(index + 1, len(data)):
            tid = pattern[1].intersection(data[i][1])
            if len(tid) >= self._minSup:
                freqPattern = pattern[0] + ' ' + data[i][0]
                freqPatterns[freqPattern] = len(tid)
                freqPatterns.update(self._genPatterns(data[i], (freqPattern, tid), data))
        return freqPatterns

    def startMine(self):
        """
        Frequent pattern mining process will start from here
        """

        self._startTime = _ab._time.time()
        conf = SparkConf().setAppName("Parallel ECLAT").setMaster("local[*]")
        sc = SparkContext(conf=conf)

        data = sc.textFile(self._iFile, self._numWorkers)\
            .map(lambda line: [int(y) for y in line.rstrip().split(self._sep)])

        frequentItems = data.flatMap(lambda trans: [(str(item), trans[0]) for item in trans[1:]]) \
            .groupByKey() \
            .filter(lambda x: len(x[1]) >= self._minSup) \
            .mapValues(set) \
            .persist()

        freqItems = dict(frequentItems.collect())
        print(len(freqItems))
        self._finalPatterns = {k: len(v) for k, v in freqItems.items()}

        freqPatterns = list(frequentItems.map(lambda x: self._genPatterns(x, x, list(freqItems.items())))
                      .filter(lambda x: len(x) != 0).collect())
        for value in freqPatterns:
            self._finalPatterns.update(value)

        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        sc.stop()

if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:
            _ap = PECLAT(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:
            _ap = PECLAT(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        _ap.startMine()
        _finalPatterns = _ap.getPatterns()
        print("Total number of Frequent Patterns:", len(_finalPatterns))
        _ap.savePatterns(_ab._sys.argv[2])
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



#  Copyright (C)  2021 Rage Uday Kiran
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.


from PAMI.uncertainPeriodicFrequentPattern.basic import abstract as _ab

_minSup = float()
_maxPer = float()
_lno = int()
_first = int()
_last = int()


class _Item:
    """
    A class used to represent the item with probability in transaction of dataset

    Attributes:
    __________
        item : int or string
            Represents the name of the item
        probability : float
            Represent the existential probability(likelihood presence) of an item
    """

    def __init__(self, item, probability):
        self.item = item
        self.probability = probability


class _Node(object):
    """
        A class used to represent the node of frequentPatternTree

        ...
        Attributes:
        ----------
            item : int
                storing item of a node
            probability : int
                To maintain the expected support of node
            parent : node
                To maintain the parent of every node
            children : list
                To maintain the children of node

        Methods:
        -------
            addChild(itemName)
                storing the children to their respective parent nodes
    """

    def __init__(self, item, children):
        self.item = item
        self.probability = 1
        self.secondProbability = 1
        self.p = 1
        self.children = children
        self.parent = None
        self.TimeStamps = []

    def addChild(self, node):
        """
            to add children details to parent node

            :param node: children node

            :return: update parent node children
        """
        self.children[node.item] = node
        node.parent = self


class _Tree(object):
    """
        A class used to represent the frequentPatternGrowth tree structure

        ...

        Attributes:
        ----------
            root: Node
                Represents the root node of the tree
            summaries: dictionary
                storing the nodes with same item name
            info: dictionary
                stores the support of items


        Methods:
        -------
            addTransaction(transaction)
                creating transaction as a branch in Tree
            addConditionalTransaction(prefixPaths, supportOfItems)
                construct the conditional tree for prefix paths
            getConditionalPatterns(Node)
                generates the conditional patterns from tree for specific node
            conditionalTransactions(prefixPaths,Support)
                takes the prefixPath of a node and support at child of the path and extract the frequent items from
                prefixPaths and generates prefixPaths with items which are frequent
            remove(Node)
                removes the node from tree once after generating all the patterns respective to the node
            generatePatterns(Node)
                starts from the root node of the tree and mines the frequent patterns

    """

    def __init__(self):
        self.root = _Node(None, {})
        self.summaries = {}
        self.info = {}

    def addTransaction(self, transaction, tid):
        """
            adding transaction into tree

            :param transaction : it represents the one transactions in database

            :type transaction : list

            :param tid : the timestamp of transaction

            :type tid : list
        """
        currentNode = self.root
        k = 0
        for i in range(len(transaction)):
            k += 1
            if transaction[i].item not in currentNode.children:
                newNode = _Node(transaction[i].item, {})
                newNode.k = k
                newNode.secondProbability = transaction[i].probability
                l1 = i - 1
                temp = []
                while l1 >= 0:
                    temp.append(transaction[l1].probability)
                    l1 -= 1
                if len(temp) == 0:
                    newNode.probability = round(transaction[i].probability, 2)
                else:
                    newNode.probability = round(max(temp) * transaction[i].probability, 2)
                currentNode.addChild(newNode)
                if transaction[i].item in self.summaries:
                    self.summaries[transaction[i].item].append(newNode)
                else:
                    self.summaries[transaction[i].item] = [newNode]
                currentNode = newNode
            else:
                currentNode = currentNode.children[transaction[i].item]
                currentNode.secondProbability = max(transaction[i].probability, currentNode.secondProbability)
                currentNode.k = k
                l1 = i - 1
                temp = []
                while l1 >= 0:
                    temp.append(transaction[l1].probability)
                    l1 -= 1
                if len(temp) == 0:
                    currentNode.probability += round(transaction[i].probability, 2)
                else:
                    nn = max(temp) * transaction[i].probability
                    currentNode.probability += round(nn, 2)
        currentNode.TimeStamps = currentNode.TimeStamps + tid

    def addConditionalPatterns(self, transaction, tid, sup):
        """
            constructing conditional tree from prefixPaths

            :param transaction : it represents the one transactions in database

            :type transaction : list

            :param tid : timestamps of a pattern or transaction in tree

            :param tid : list

            :param sup : support of prefixPath taken at last child of the path

            :type sup : int
        """
        currentNode = self.root
        k = 0
        for i in range(len(transaction)):
            k += 1
            if transaction[i] not in currentNode.children:
                newNode = _Node(transaction[i], {})
                newNode.k = k
                newNode.probability = sup
                currentNode.addChild(newNode)
                if transaction[i] in self.summaries:
                    self.summaries[transaction[i]].append(newNode)
                else:
                    self.summaries[transaction[i]] = [newNode]
                currentNode = newNode
            else:
                currentNode = currentNode.children[transaction[i]]
                currentNode.k = k
                currentNode.probability += sup
        currentNode.TimeStamps = currentNode.TimeStamps + tid

    def conditionalPatterns(self, alpha):
        """generates all the conditional patterns of respective node

                :param alpha : it represents the Node in tree

                :type alpha : Node
        """
        finalPatterns = []
        finalSets = []
        sup = []
        for i in self.summaries[alpha]:
            set1 = i.TimeStamps
            s = i.probability
            set2 = []
            while i.parent.item is not None:
                set2.append(i.parent.item)
                i = i.parent
            if len(set2) > 0:
                set2.reverse()
                finalPatterns.append(set2)
                finalSets.append(set1)
                sup.append(s)
        finalPatterns, finalSets, support, info = self.conditionalTransactions(finalPatterns, finalSets, sup)
        return finalPatterns, finalSets, support, info

    def removeNode(self, nodeValue):
        """removing the node from tree

            :param nodeValue : it represents the node in tree

            :type nodeValue : node
        """
        for i in self.summaries[nodeValue]:
            i.parent.TimeStamps = i.parent.TimeStamps + i.TimeStamps
            del i.parent.children[nodeValue]

    def getPeriodAndSupport(self, support, TimeStamps):
        """

        Parameters
        ----------
        support: support of pattern
        TimeStamps: timmeStamps of a pattern

        Returns
        -------
        support and period

        """
        global _maxPer
        global _lno
        TimeStamps.sort()
        cur = 0
        per = 0
        sup = support
        for j in range(len(TimeStamps)):
            per = max(per, TimeStamps[j] - cur)
            if per > _maxPer:
                return [0, 0]
            cur = TimeStamps[j]
        per = max(per, _lno - cur)
        return [sup, per]

    def conditionalTransactions(self, conditionalPatterns, conditionalTimeStamps, support):
        """ It generates the conditional patterns with frequent items

            :param conditionalPatterns : conditional patterns generated from conditionalPatterns() method for respective node

            :type conditionalPatterns : list

            :param conditionalTimeStamps : timestamps of respective conditional timestamps

            :type conditionalTimeStamps : list

            :param support : the support of conditional pattern in tree

            :type support : list
        """
        global _minSup, _maxPer, _lno
        pat = []
        TimeStamps = []
        sup = []
        data1 = {}
        count = {}
        for i in range(len(conditionalPatterns)):
            for j in conditionalPatterns[i]:
                if j in data1:
                    data1[j] = data1[j] + conditionalTimeStamps[i]
                    count[j] += support[i]
                else:
                    data1[j] = conditionalTimeStamps[i]
                    count[j] = support[i]
        updatedDict = {}
        for m in data1:
            updatedDict[m] = self.getPeriodAndSupport(count[m], data1[m])
        updatedDict = {k: v for k, v in updatedDict.items() if v[0] >= _minSup and v[1] <= _maxPer}
        count = 0
        for p in conditionalPatterns:
            p1 = [v for v in p if v in updatedDict]
            trans = sorted(p1, key=lambda x: (updatedDict.get(x)[0]), reverse=True)
            if len(trans) > 0:
                pat.append(trans)
                TimeStamps.append(conditionalTimeStamps[count])
                sup.append(support[count])
            count += 1
        return pat, TimeStamps, sup, updatedDict

    def generatePatterns(self, prefix, periodic):
        """generates the patterns

            :param prefix : forms the combination of items

            :type prefix : list
        """
        global _minSup
        for i in sorted(self.summaries, key=lambda x: (self.info.get(x)[0])):
            pattern = prefix[:]
            pattern.append(i)
            s = 0
            for x in self.summaries[i]:
                if x.k <= 2:
                    s += x.probability
                elif x.k >= 3:
                    n = x.probability * pow(x.secondProbability, (x.k - 2))
                    s += n
            periodic[tuple(pattern)] = self.info[i]
            if s >= _minSup:
                patterns, TimeStamps, support, info = self.conditionalPatterns(i)
                conditionalTree = _Tree()
                conditionalTree.info = info.copy()
                for pat in range(len(patterns)):
                    conditionalTree.addConditionalPatterns(patterns[pat], TimeStamps[pat], support[pat])
                if len(patterns) > 0:
                    conditionalTree.generatePatterns(pattern, periodic)
            self.removeNode(i)


class PTubeP(_ab._periodicFrequentPatterns):
    """
        Periodic-TubeP is  to discover periodic-frequent patterns in a temporal database.

        Reference:
        --------


        Attributes:
        ----------
            iFile: file
                Name of the Input file or path of input file
            oFile: file
                Name of the output file or path of output file
            minSup: int or float or str
                The user can specify minSup either in count or proportion of database size.
                If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
                Otherwise, it will be treated as float.
                Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
            maxPer: int or float or str
                The user can specify maxPer either in count or proportion of database size.
                If the program detects the data type of maxPer is integer, then it treats maxPer is expressed in count.
                Otherwise, it will be treated as float.
                Example: maxPer=10 will be treated as integer, while maxPer=10.0 will be treated as float
            sep: str
                This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
                However, the users can override their default separator.
            memoryUSS: float
                To store the total amount of USS memory consumed by the program
            memoryRSS: float
                To store the total amount of RSS memory consumed by the program
            startTime: float
                To record the start time of the mining process
            endTime: float
                To record the completion time of the mining process
            Database: list
                To store the transactions of a database in list
            mapSupport: Dictionary
                To maintain the information of item and their frequency
            lno: int
                To represent the total no of transaction
            tree: class
                To represents the Tree class
            itemSetCount: int
                To represents the total no of patterns
            finalPatterns: dict
                To store the complete patterns
        Methods:
        -------
            startMine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            savePatterns(oFile)
                Complete set of periodic-frequent patterns will be loaded in to a output file
            getPatternsAsDataFrame()
                Complete set of periodic-frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function
            creatingItemSets(fileName)
                Scans the dataset and stores in a list format
            updateDatabases()
                Update the database by removing aperiodic items and sort the Database by item decreased support
            buildTree()
                After updating the Database, remaining items will be added into the tree by setting root node as null
            convert()
                to convert the user specified value
            PeriodicFrequentOneItems()
                To extract the one-length periodic-frequent items

        Executing the code on terminal:
        -------
            Format:
            ------
                python3 PTubeP.py <inputFile> <outputFile> <minSup> <maxPer>

            Examples:
            --------
                python3 PTubeP.py sampleTDB.txt patterns.txt 0.3 4     (minSup and maxPer will be considered in support count or frequency)

        Sample run of importing the code:
        -------------------

            from PAMI.uncertainPeriodicFrequentPattern.basic import PTubeP as alg

            obj = alg.PTubeP(iFile, minSup, maxPer)

            obj.startMine()

            periodicFrequentPatterns = obj.getPatterns()

            print("Total number of Periodic Frequent Patterns:", len(periodicFrequentPatterns))

            obj.savePatterns(oFile)

            Df = obj.getPatternsAsDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

        Credits:
        -------
            The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.\n

    """
    _startTime = float()
    _endTime = float()
    _minSup = float()
    _maxPer = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _rank = {}
    _lno = 0
    _periodic = {}

    def _creatingItemSets(self):
        """
            Storing the complete transactions of the database/input file in a database variable


        """

        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            uncertain, data, ts = [], [], []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'TS' in i:
                ts = self._iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self._iFile['Transactions'].tolist()
            if 'uncertain' in i:
                uncertain = self._iFile['uncertain'].tolist()
            for k in range(len(data)):
                tr = [ts[k]]
                for j in range(len(k)):
                    product = _Item(data[k][j], uncertain[k][j])
                    tr.append(product)
                self._Database.append(tr)
                self._lno += 1

            # print(self.Database)
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    tr = []
                    for i in temp[1:]:
                        i1 = i.index('(')
                        i2 = i.index(')')
                        item = i[0:i1]
                        probability = float(i[i1 + 1:i2])
                        product = _Item(item, probability)
                        tr.append(product)
                    self._lno += 1
                    self._Database.append(temp)
            else:
                try:
                    count = 0
                    with open(self._iFile, 'r') as f:
                        for line in f:
                            count += 1
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            tr = [count]
                            for i in temp:
                                i1 = i.index('(')
                                i2 = i.index(')')
                                item = i[0:i1]
                                probability = float(i[i1 + 1:i2])
                                product = _Item(item, probability)
                                tr.append(product)
                            self._lno += 1
                            self._Database.append(tr)
                except IOError:
                    print("File Not Found")

    def _PeriodicFrequentOneItems(self):
        """takes the transactions and calculates the support of each item in the dataset and assign the
                            ranks to the items by decreasing support and returns the frequent items list

        """
        global first, last
        mapSupport = {}
        for i in self._Database:
            n = int(i[0])
            for j in i[1:]:
                if j.item not in mapSupport:
                    mapSupport[j.item] = [round(j.probability, 2), abs(0 - n), n]
                else:
                    mapSupport[j.item][0] += round(j.probability, 2)
                    mapSupport[j.item][1] = max(mapSupport[j.item][1], abs(n - mapSupport[j.item][2]))
                    mapSupport[j.item][2] = n
        for key in mapSupport:
            mapSupport[key][1] = max(mapSupport[key][1], self._lno - mapSupport[key][2])
        mapSupport = {k: [round(v[0], 2), v[1]] for k, v in mapSupport.items() if
                      v[1] <= self._maxPer and v[0] >= self._minSup}
        plist = [k for k, v in sorted(mapSupport.items(), key=lambda x: (x[1][0], x[0]), reverse=True)]
        self._rank = dict([(index, item) for (item, index) in enumerate(plist)])
        return mapSupport, plist

    def _buildTree(self, data, info):
        """it takes the transactions and support of each item and construct the main tree with setting root
                            node as null

            :param data : it represents the one transactions in database

            :type data : list

            :param info : it represents the support of each item

            :type info : dictionary
        """
        rootNode = _Tree()
        rootNode.info = info.copy()
        for i in range(len(data)):
            set1 = [data[i][0]]
            rootNode.addTransaction(data[i][1:], set1)
        return rootNode

    def _updateTransactions(self, dict1):
        """remove the items which are not frequent from transactions and updates the transactions with rank of items

                :param dict1 : frequent items with support

                :type dict1 : dictionary
        """
        list1 = []
        for tr in self._Database:
            list2 = [int(tr[0])]
            for i in range(1, len(tr)):
                if tr[i].item in dict1:
                    list2.append(tr[i])
            if len(list2) >= 2:
                basket = list2[1:]
                basket.sort(key=lambda val: self._rank[val.item])
                list2[1:] = basket[0:]
                list1.append(list2)
        return list1

    def _Check(self, i, x):
        """To check the presence of item or pattern in transaction

            :param x: it represents the pattern

            :type x : list

            :param i : represents the uncertain transactions

            :type i : list
        """
        for m in x:
            k = 0
            for n in i:
                if m == n.item:
                    k += 1
            if k == 0:
                return 0
        return 1

    def _convert(self, value):
        """
            To convert the given user specified value

            :param value: user specified value

            :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = int(len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = int(len(self._Database) * value)
            else:
                value = int(value)

        return value

    def _removeFalsePositives(self):
        """
        To remove false positives in generated patterns

        :return: original patterns
        """
        periods = {}
        for i in self._Database:
            for x, y in self._periodic.items():
                if len(x) == 1:
                    periods[x] = y
                else:
                    s = 1
                    check = self._Check(i[1:], x)
                    if check == 1:
                        for j in i[1:]:
                            if j.item in x:
                                s *= j.probability
                        if x in periods:
                            periods[x][0] += s
                        else:
                            periods[x] = [s, y[1]]
        for x, y in periods.items():
            if y[0] >= _minSup:
                sample = str()
                for i in x:
                    sample = sample + i + " "
                self._finalPatterns[sample] = y

    def startMine(self):
        """Main method where the patterns are mined by constructing tree and remove the remove the false patterns
                           by counting the original support of a patterns


               """
        global _minSup, _maxPer, _first, _last, _lno
        self._startTime = _ab._time.time()
        self._creatingItemSets()
        self._minSup = self._convert(self._minSup)
        self._maxPer = self._convert(self._maxPer)
        self._finalPatterns = {}
        _minSup, _maxPer, _lno = self._minSup, self._maxPer, len(self._Database)
        mapSupport, plist = self._PeriodicFrequentOneItems()
        updatedTrans = self._updateTransactions(mapSupport)
        info = {k: v for k, v in mapSupport.items()}
        root = self._buildTree(updatedTrans, info)
        self._periodic = {}
        root.generatePatterns([], self._periodic)
        self._removeFalsePositives()
        print("Periodic Frequent patterns were generated successfully using Periodic-TubeP algorithm")
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process

        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process

        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process

        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe

        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b[0], b[1]])
            dataframe = _ab._pd.DataFrame(data, columns=['Patterns', 'Support', 'Periodicity'])
        return dataframe

    def savePatterns(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file

        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:
            _ap = PTubeP(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:
            _ap = PTubeP(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        _ap.startMine()
        _Patterns = _ap.getPatterns()
        print("Total number of Patterns:", len(_Patterns))
        _ap.savePatterns(_ab._sys.argv[2])
        # print(ap.getPatternsAsDataFrame())
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
    else:
        l = [140]
        for i in l:
            ap = PTubeP('/home/apiiit-rkv/Desktop/uncertain/congestion', i, 2000, ' ')
            ap.startMine()
            Patterns = ap.getPatterns()
            print("Total number of Patterns:", len(Patterns))
            ap.savePatterns('/home/apiiit-rkv/Desktop/uncertain/output')
            memUSS = ap.getMemoryUSS()
            print("Total Memory in USS:", memUSS)
            memRSS = ap.getMemoryRSS()
            print("Total Memory in RSS", memRSS)
            run = ap.getRuntime()
            print("Total ExecutionTime in ms:", run)
        print("Error! The number of input parameters do not match the total number of parameters provided")



import pandas as pd


def generateLatexCode(result):
    latexwriter = open("Latexfile.tex", "w")
    latexwriter.write("")
    plots = int(input("How many lines do you want to plot: "))
    x_label = input("Enter xlable: ")
    y_label = input("Enter ylable: ")
    x_min = int(input("Enter xmin: "))
    x_max = int(input("Enter x_max: "))
    axis_cs = input("Enter axis cs separated by comma: ")
    axis_cs = axis_cs.split(",")
    latexwriter = open("Latexfile.tex", "a")
    latexwriter.write("\\begin{axis}[\n\txlabel={\\Huge{" + x_label + "}},")
    latexwriter.write("\n\tylabel={\\Huge{" + y_label + "}},")
    latexwriter.write("\n\txmin=" + str(x_min) + ", xmax=" + str(x_max) + ",")
    latexwriter.write("\n\tlegend style={at={(axis cs: " + str(axis_cs[0]) + "," + str(axis_cs[1]) + ")}},\n]")
    for i in range(0, plots):
        legendary = input("Enter legendary title for plot: ")
        print("Enter the color for " + legendary + ": ", end="")
        color = input()
        algo_name = input("Plot results for which algorithm? (Please enter algorithm name):")
        x_axis = input("Which column will become x axis data from dataframe: ")
        y_axis = input("Which column will become y axis data from dataframe: ")
        #xaxis = result[x_axis].values.tolist()
        #yaxis = result[y_axis].values.tolist()
        df2 = result[['minSup','runtime']] [result['algorithm'] == algo_name]
        print(df2)
        algorithm = df2.values.tolist()
        print(algorithm)

        latexwriter.write("\n\\addplot+  [" + color + "]\n\tcoordinates {\n")

        for i in range(0, len(algorithm)):
            latexwriter.write("(" + str(algorithm[i][0]) + ","+ str(algorithm[i][1]) + ")\n")
        latexwriter.write("\t};   \\addlegendentry{" + legendary + "}\n")
    latexwriter.write("\\end{axis}")
    print("Latex file generated successfully")

if __name__ == "__main__":


    data = {'Name': ['Jai', 'Princi', 'Gaurav', 'Anuj'],
            'Age': [27, 24, 22, 32],
            'Address': [0, 1, 2, 3],
            'Qualification': [8, 9, 10, 11]}
    data1 = pd.DataFrame(data)
    print(data1)
    #print(data1['Name'].values.tolist())
    generateLatexCode(data1)



import os
import csv
import time
import numpy as np
import pycuda.gpuarray as gpuarray
import pycuda.autoinit
import psutil


class cudaEclatGCT:
    __time = 0
    __memRSS = 0
    __memUSS = 0
    __GPU_MEM = 0
    filePath = ""
    sep = ""
    minSup = 0
    Patterns = {}

    def __init__(self, filePath, sep, minSup):
        self.filePath = filePath
        self.sep = sep
        self.minSup = minSup
        self.__time = 0
        self.__memRSS = 0
        self.__memUSS = 0

    def read_data(self, data_path, sep):
        data = []
        if not os.path.isfile(data_path):
            raise ValueError('Invalid data path.' + data_path)
        with open(data_path, 'r') as f:
            file = csv.reader(f, delimiter=sep, quotechar='\r')
            lineNo = 1
            for row in file:
                data.append([str(item) for item in row if item != ''])
                lineNo += 1
        return data, lineNo

    def write_result(self, result, result_path):
        file = open(result_path, 'w')
        for itemset, support in result.items():
            file.write(str(itemset) + ' : ' + str(support) + '\n')
        file.close()

    def compute_vertical_bitvector_data(self, data):
        #---build item to idx mapping---#
        idx = 0
        item2idx = {}
        for transaction in data:
            for item in transaction:
                if not item in item2idx:
                    item2idx[item] = idx
                    idx += 1
        idx2item = {idx: str(int(item)) for item, idx in item2idx.items()}
        #---build vertical data---#
        vb_data = np.zeros((len(item2idx), len(data)), dtype=np.uint16)
        for trans_id, transaction in enumerate(data):
            for item in transaction:
                vb_data[item2idx[item], trans_id] = 1
        vb_data = gpuarray.to_gpu(vb_data.astype(np.uint16))
        return vb_data, idx2item

    def get_time(self):
        return self.__time

    def get_memRSS(self):
        return self.__memRSS

    def get_memUSS(self):
        return self.__memUSS

    def get_GPU_MEM(self):
        return self.__GPU_MEM

    def get_Patterns(self):
        return self.Patterns

    def get_numberOfPatterns(self):
        return len(self.Patterns)

    def eclat(self, basePattern, final, vb_data, idx2item, item2idx):
        newBasePattern = []
        for i in range(0, len(basePattern)):
            item1 = basePattern[i]
            i1_list = item1.split()
            for j in range(i + 1, len(basePattern)):
                item2 = basePattern[j]
                i2_list = item2.split()
                if i1_list[:-1] == i2_list[:-1]:
                    unionOfKey = list(set(i1_list) | set(i2_list))
                    unionOfKey.sort()
                    valueList = []
                    for key in unionOfKey:
                        valueList.append(item2idx[key])
                    total = vb_data[valueList[0]]
                    for k in range(1, len(valueList)):
                        total = total.__mul__(vb_data[valueList[k]])
                    support = gpuarray.sum(total).get()
                    if support >= self.minSup:
                        newBasePattern.append(" ".join(unionOfKey))
                        final[" ".join(unionOfKey)] = support

        if len(newBasePattern) > 0:
            self.eclat(newBasePattern, final, vb_data, idx2item, item2idx)

    def startMine(self):
        startTime = time.time()
        basePattern = []
        final = {}

        data, lineNo = self.read_data(self.filePath, self.sep)
        vb_data, idx2item = self.compute_vertical_bitvector_data(data)
        if self.minSup < 1:
            self.minSup = int(lineNo * self.minSup)

        for i in range(len(vb_data)):
            if gpuarray.sum(vb_data[i]).get() >= self.minSup:
                basePattern.append(idx2item[i])
                final[idx2item[i]] = gpuarray.sum(vb_data[i]).get()

        # reverse idx2item
        item2idx = {idx2item[i]: i for i in idx2item}
        self.eclat(basePattern, final, vb_data, idx2item, item2idx)
        self.__time = time.time() - startTime
        self.__memRSS = psutil.Process(os.getpid()).memory_info().rss
        self.__memUSS = psutil.Process(os.getpid()).memory_full_info().uss
        self.Patterns = final
        self.__GPU_MEM = vb_data.nbytes


if __name__ == "__main__":
    filePath = "datasets\\transactional_T10I4D100K.csv"
    # filePath = "file.txt"
    sep = "\t"
    support = 500
    cudaEclatGCT = cudaEclatGCT(filePath, sep, support)
    cudaEclatGCT.startMine()
    print("Time: ", cudaEclatGCT.get_time())
    print("Memory RSS: ", cudaEclatGCT.get_memRSS())
    print("Memory USS: ", cudaEclatGCT.get_memUSS())
    print("GPU MEM: ", cudaEclatGCT.get_GPU_MEM())
    print("Patterns: ", cudaEclatGCT.get_numberOfPatterns())



import statistics
import validators
from urllib.request import urlopen
import pandas as pd

class utilityDatabaseStats:
    """
    utilityDatabaseStats is class to get stats of database.

        Attributes:
        ----------
        inputFile : file
            input file path
        database : dict
            store time stamp and its transaction
        lengthList : list
            store length of all transaction
        utility : dict
            store utility each item
        sep : str
            separator in file. Default is tab space.

        Methods:
        -------
        run()
            execute readDatabase function
        readDatabase()
            read database from input file
        getDatabaseSize()
            get the size of database
        getMinimumTransactionLength()
            get the minimum transaction length
        getAverageTransactionLength()
            get the average transaction length. It is sum of all transaction length divided by database length.
        getMaximumTransactionLength()
            get the maximum transaction length
        getStandardDeviationTransactionLength()
            get the standard deviation of transaction length
        getSortedListOfItemFrequencies()
            get sorted list of item frequencies
        getSortedListOfTransactionLength()
            get sorted list of transaction length
        save(data, outputFile)
            store data into outputFile
        getMinimumUtility()
            get the minimum utility
        getAverageUtility()
            get the average utility
        getMaximumUtility()
            get the maximum utility
        getSortedUtilityValuesOfItem()
            get sorted utility values each item
    """
    def __init__(self, inputFile, sep='\t'):
        """
        :param inputFile: input file name or path
        :type inputFile: str
        """
        self.inputFile = inputFile
        self.database = {}
        self.lengthList = []
        self.utility = {}
        self.sep = sep

    def run(self):
        self.readDatabase()

    def creatingItemSets(self):
        """
            Storing the complete transactions of the database/input file in a database variable


        """
        self.Database = []
        self.utilityValues = []
        if isinstance(self.inputFile, pd.DataFrame):
            if self.inputFile.empty:
                print("its empty..")
            i = self.inputFile.columns.values.tolist()
            if 'Transactions' in i:
                self.Database = self.inputFile['Transactions'].tolist()
            if 'Patterns' in i:
                self.Database = self.inputFile['Patterns'].tolist()
            if 'Utility' in i:
                self.utilityValues = self.inputFile['Utility'].tolist()

        if isinstance(self.inputFile, str):
            if validators.url(self.inputFile):
                data = urlopen(self.inputFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(":")]
                    transaction = [s for s in temp[0].split(self.sep)]
                    self.Database.append([x for x in transaction if x])
                    utilities = [int(s) for s in temp[2].split(self.sep)]
                    self.utilityValues.append([x for x in utilities if x])
            else:
                try:
                    with open(self.inputFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(":")]
                            transaction = [s for s in temp[0].split(self.sep)]
                            self.Database.append([x for x in transaction if x])
                            utilities = [int(s) for s in temp[2].split(self.sep)]
                            self.utilityValues.append([x for x in utilities if x])
                except IOError:
                    print("File Not Found")
                    quit()

    def readDatabase(self):
        """
        read database from input file and store into database and size of each transaction.
        """
        numberOfTransaction = 0
        self.creatingItemSets()
        for k in range(len(self.Database)):
            numberOfTransaction += 1
            transaction = self.Database[k]
            utilities = self.utilityValues[k]
            self.database[numberOfTransaction] = transaction
            for i in range(len(transaction)):
                self.utility[transaction[i]] = self.utility.get(transaction[i],0)
                self.utility[transaction[i]] += utilities[i]
        self.lengthList = [len(s) for s in self.database.values()]
        self.utility = {k: v for k, v in sorted(self.utility.items(), key=lambda x:x[1], reverse=True)}

    def getDatabaseSize(self):
        """
        get the size of database
        :return: data base size
        """
        return len(self.database)

    def getTotalNumberOfItems(self):
        """
        get the number of items in database.
        :return: number of items
        """
        return len(self.getSortedListOfItemFrequencies())

    def getMinimumTransactionLength(self):
        """
        get the minimum transaction length
        :return: minimum transaction length
        """
        return min(self.lengthList)

    def getAverageTransactionLength(self):
        """
        get the average transaction length. It is sum of all transaction length divided by database length.
        :return: average transaction length
        """
        totalLength = sum(self.lengthList)
        return totalLength / len(self.database)

    def getMaximumTransactionLength(self):
        """
        get the maximum transaction length
        :return: maximum transaction length
        """
        return max(self.lengthList)

    def getStandardDeviationTransactionLength(self):
        """
        get the standard deviation transaction length
        :return: standard deviation transaction length
        """
        return statistics.pstdev(self.lengthList)

    def getVarianceTransactionLength(self):
        """
        get the variance transaction length
        :return: variance transaction length
        """
        return statistics.variance(self.lengthList)

    def getNumberOfItems(self):
        """
        get the number of items in database.
        :return: number of items
        """
        return len(self.getSortedListOfItemFrequencies())

    def getSparsity(self):
        # percentage of 0 dense dataframe
        """
        get the sparsity of database
        :return: database sparsity
        """
        matrixSize = self.getDatabaseSize()*len(self.getSortedListOfItemFrequencies())
        return (matrixSize - sum(self.getSortedListOfItemFrequencies().values())) / matrixSize

    def getSortedListOfItemFrequencies(self):
        """
        get sorted list of item frequencies
        :return: item frequencies
        """
        itemFrequencies = {}
        for tid in self.database:
            for item in self.database[tid]:
                itemFrequencies[item] = itemFrequencies.get(item, 0)
                itemFrequencies[item] += 1
        return {k: v for k, v in sorted(itemFrequencies.items(), key=lambda x:x[1], reverse=True)}

    def getTransanctionalLengthDistribution(self):
        """
        get transaction length
        :return: transaction length
        """
        transactionLength = {}
        for length in self.lengthList:
            transactionLength[length] = transactionLength.get(length, 0)
            transactionLength[length] += 1
        return {k: v for k, v in sorted(transactionLength.items(), key=lambda x:x[0])}

    def save(self, data, outputFile):
        """
        store data into outputFile
        :param data: input data
        :type data: dict
        :param outputFile: output file name or path to store
        :type outputFile: str
        """
        with open(outputFile, 'w') as f:
            for key, value in data.items():
                f.write(f'{key}\t{value}\n')

    def getTotalUtility(self):
        """
        get sum of utility
        :return: total utility
        """
        return sum(list(self.utility.values()))

    def getMinimumUtility(self):
        """
        get the minimum utility
        :return: minimum utility
        """
        return min(list(self.utility.values()))

    def getAverageUtility(self):
        """
        get the average utility
        :return: average utility
        """
        return sum(list(self.utility.values())) / len(self.utility)

    def getMaximumUtility(self):
        """
        get the maximum utility
        :return: maximum utility
        """
        return max(list(self.utility.values()))

    def getSortedUtilityValuesOfItem(self):
        """
        get sorted utility value each item. key is item and value is utility of item
        :return: sorted dictionary utility value of item
        """
        return self.utility


if __name__ == '__main__':
    data = {'ts': [1, 1, 3, 4, 5, 6, 7],

            'Transactions': [['a', 'd', 'e'], ['b', 'a', 'f', 'g', 'h'], ['b', 'a', 'd', 'f'], ['b', 'a', 'c'],
                             ['a', 'd', 'g', 'k'],

                             ['b', 'd', 'g', 'c', 'i'], ['b', 'd', 'g', 'e', 'j']]}

    data = pd.DataFrame.from_dict(data)
    import PAMI.extras.dbStats.utilityDatabaseStats as uds
    import PAMI.extras.graph.plotLineGraphFromDictionary as plt

    obj = utilityDatabaseStats(data)
    obj = uds.utilityDatabaseStats('/Users/Likhitha/Downloads/internalUtility_accidents.csv', sep=',')
    obj.run()
    print(f'Database size : {obj.getDatabaseSize()}')
    print(f'Minimum Transaction Size : {obj.getMinimumTransactionLength()}')
    print(f'Average Transaction Size : {obj.getAverageTransactionLength()}')
    print(f'Maximum Transaction Size : {obj.getMaximumTransactionLength()}')
    print(f'Standard Deviation Transaction Size : {obj.getStandardDeviationTransactionLength()}')
    print(f'Variance : {obj.getVarianceTransactionLength()}')
    print(f'Sparsity : {obj.getSparsity()}')
    print(f'Number of items : {obj.getTotalNumberOfItems()}')
    print(f'Minimum utility : {obj.getMinimumUtility()}')
    print(f'Average utility : {obj.getAverageUtility()}')
    print(f'Maximum utility : {obj.getMaximumUtility()}')
    print(f'sorted utility value each item : {obj.getSortedUtilityValuesOfItem()}')
    itemFrequencies = obj.getSortedListOfItemFrequencies()
    transactionLength = obj.getTransanctionalLengthDistribution()
    numberOfTransactionPerTimeStamp = obj.getNumberOfTransactionsPerTimestamp()
    plt.plotLineGraphFromDictionary(itemFrequencies, 100, 'itemFrequencies', 'item rank', 'frequency')
    plt.plotLineGraphFromDictionary(transactionLength, 100, 'transaction length', 'transaction length', 'frequency')
    plt.plotLineGraphFromDictionary(numberOfTransactionPerTimeStamp, 100)




class predictedClass2Transaction:
    def getBinaryTransaction(self,predicted_classes,minThreshold=0.8):
        self.predictions_dict ={}
        for name, p, box in predicted_classes:
            if p > minThreshold :
                    if name not in self.predictions_dict:
                        self.predictions_dict[name] = [p, ]
                    else:
                        self.predictions_dict[name].append(p)
        return self.predictions_dict
